# Copyright 2021 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import multiprocessing
import time
from datetime import datetime, timedelta
from multiprocessing import Process
from sys import platform
from typing import List

import pandas as pd
import pytest
from _pytest.nodes import Item

from feast import FeatureStore
from tests.data.data_creator import create_dataset
from tests.integration.feature_repos.integration_test_repo_config import (
    IntegrationTestRepoConfig,
)
from tests.integration.feature_repos.repo_configuration import (
    FULL_REPO_CONFIGS,
    REDIS_CLUSTER_CONFIG,
    REDIS_CONFIG,
    Environment,
    TestData,
    construct_test_environment,
    construct_universal_test_data,
)

logger = logging.getLogger(__name__)


def pytest_configure(config):
    if platform in ["darwin", "windows"]:
        multiprocessing.set_start_method("spawn")
    else:
        multiprocessing.set_start_method("fork")
    config.addinivalue_line(
        "markers", "integration: mark test that has external dependencies"
    )
    config.addinivalue_line("markers", "benchmark: mark benchmarking tests")
    config.addinivalue_line(
        "markers", "universal: mark tests that use the universal feature repo"
    )


def pytest_addoption(parser):
    parser.addoption(
        "--integration",
        action="store_true",
        default=False,
        help="Run tests with external dependencies",
    )
    parser.addoption(
        "--benchmark", action="store_true", default=False, help="Run benchmark tests",
    )
    parser.addoption(
        "--universal", action="store_true", default=False, help="Run universal tests",
    )


def pytest_collection_modifyitems(config, items: List[Item]):
    should_run_integration = config.getoption("--integration") is True
    should_run_benchmark = config.getoption("--benchmark") is True
    should_run_universal = config.getoption("--universal") is True

    integration_tests = [t for t in items if "integration" in t.keywords]
    if not should_run_integration:
        for t in integration_tests:
            items.remove(t)
    else:
        items.clear()
        for t in integration_tests:
            items.append(t)

    benchmark_tests = [t for t in items if "benchmark" in t.keywords]
    if not should_run_benchmark:
        for t in benchmark_tests:
            items.remove(t)
    else:
        items.clear()
        for t in benchmark_tests:
            items.append(t)

    universal_tests = [t for t in items if "universal" in t.keywords]
    if should_run_universal:
        items.clear()
        for t in universal_tests:
            items.append(t)


@pytest.fixture
def simple_dataset_1() -> pd.DataFrame:
    now = datetime.utcnow()
    ts = pd.Timestamp(now).round("ms")
    data = {
        "id_join_key": [1, 2, 1, 3, 3],
        "float_col": [0.1, 0.2, 0.3, 4, 5],
        "int64_col": [1, 2, 3, 4, 5],
        "string_col": ["a", "b", "c", "d", "e"],
        "ts_1": [
            ts,
            ts - timedelta(hours=4),
            ts - timedelta(hours=3),
            ts - timedelta(hours=2),
            ts - timedelta(hours=1),
        ],
    }
    return pd.DataFrame.from_dict(data)


@pytest.fixture
def simple_dataset_2() -> pd.DataFrame:
    now = datetime.utcnow()
    ts = pd.Timestamp(now).round("ms")
    data = {
        "id_join_key": ["a", "b", "c", "d", "e"],
        "float_col": [0.1, 0.2, 0.3, 4, 5],
        "int64_col": [1, 2, 3, 4, 5],
        "string_col": ["a", "b", "c", "d", "e"],
        "ts_1": [
            ts,
            ts - timedelta(hours=4),
            ts - timedelta(hours=3),
            ts - timedelta(hours=2),
            ts - timedelta(hours=1),
        ],
    }
    return pd.DataFrame.from_dict(data)


def start_test_local_server(repo_path: str, port: int):
    fs = FeatureStore(repo_path)
    fs.serve("localhost", port, no_access_log=True)


@pytest.fixture(
    params=FULL_REPO_CONFIGS, scope="session", ids=[str(c) for c in FULL_REPO_CONFIGS]
)
def environment(request, worker_id: str):
    e = construct_test_environment(request.param, worker_id=worker_id)
    proc = Process(
        target=start_test_local_server,
        args=(e.feature_store.repo_path, e.get_local_server_port()),
        daemon=True,
    )
    if e.python_feature_server and e.test_repo_config.provider == "local":
        proc.start()
        # Wait for server to start
        time.sleep(3)

    def cleanup():
        e.feature_store.teardown()
        if proc.is_alive():
            proc.kill()

    request.addfinalizer(cleanup)

    return e


@pytest.fixture(
    params=[REDIS_CONFIG, REDIS_CLUSTER_CONFIG],
    scope="session",
    ids=[str(c) for c in [REDIS_CONFIG, REDIS_CLUSTER_CONFIG]],
)
def local_redis_environment(request, worker_id):
    e = construct_test_environment(
        IntegrationTestRepoConfig(online_store=request.param), worker_id=worker_id
    )

    def cleanup():
        e.feature_store.teardown()

    request.addfinalizer(cleanup)
    return e


@pytest.fixture(scope="session")
def universal_data_sources(request, environment) -> TestData:
    def cleanup():
        # logger.info("Running cleanup in %s, Request: %s", worker_id, request.param)
        environment.data_source_creator.teardown()

    request.addfinalizer(cleanup)
    return construct_universal_test_data(environment)


@pytest.fixture(scope="session")
def redis_universal_data_sources(request, local_redis_environment):
    def cleanup():
        # logger.info("Running cleanup in %s, Request: %s", worker_id, request.param)
        local_redis_environment.data_source_creator.teardown()

    request.addfinalizer(cleanup)
    return construct_universal_test_data(local_redis_environment)


@pytest.fixture(scope="session")
def e2e_data_sources(environment: Environment, request):
    df = create_dataset()
    data_source = environment.data_source_creator.create_data_source(
        df, environment.feature_store.project, field_mapping={"ts_1": "ts"},
    )

    def cleanup():
        environment.data_source_creator.teardown()

    request.addfinalizer(cleanup)

    return df, data_source



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: feast/core/InfraObject.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from feast.protos.feast.core import DatastoreTable_pb2 as feast_dot_core_dot_DatastoreTable__pb2
from feast.protos.feast.core import DynamoDBTable_pb2 as feast_dot_core_dot_DynamoDBTable__pb2
from feast.protos.feast.core import SqliteTable_pb2 as feast_dot_core_dot_SqliteTable__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='feast/core/InfraObject.proto',
  package='feast.core',
  syntax='proto3',
  serialized_options=b'\n\020feast.proto.coreB\020InfraObjectProtoZ3github.com/feast-dev/feast/sdk/go/protos/feast/core',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1c\x66\x65\x61st/core/InfraObject.proto\x12\nfeast.core\x1a\x1f\x66\x65\x61st/core/DatastoreTable.proto\x1a\x1e\x66\x65\x61st/core/DynamoDBTable.proto\x1a\x1c\x66\x65\x61st/core/SqliteTable.proto\"7\n\x05Infra\x12.\n\rinfra_objects\x18\x01 \x03(\x0b\x32\x17.feast.core.InfraObject\"\xb6\x02\n\x0bInfraObject\x12\x1f\n\x17infra_object_class_type\x18\x01 \x01(\t\x12\x33\n\x0e\x64ynamodb_table\x18\x02 \x01(\x0b\x32\x19.feast.core.DynamoDBTableH\x00\x12\x35\n\x0f\x64\x61tastore_table\x18\x03 \x01(\x0b\x32\x1a.feast.core.DatastoreTableH\x00\x12/\n\x0csqlite_table\x18\x04 \x01(\x0b\x32\x17.feast.core.SqliteTableH\x00\x12;\n\x0c\x63ustom_infra\x18\x64 \x01(\x0b\x32#.feast.core.InfraObject.CustomInfraH\x00\x1a\x1c\n\x0b\x43ustomInfra\x12\r\n\x05\x66ield\x18\x01 \x01(\x0c\x42\x0e\n\x0cinfra_objectBY\n\x10\x66\x65\x61st.proto.coreB\x10InfraObjectProtoZ3github.com/feast-dev/feast/sdk/go/protos/feast/coreb\x06proto3'
  ,
  dependencies=[feast_dot_core_dot_DatastoreTable__pb2.DESCRIPTOR,feast_dot_core_dot_DynamoDBTable__pb2.DESCRIPTOR,feast_dot_core_dot_SqliteTable__pb2.DESCRIPTOR,])




_INFRA = _descriptor.Descriptor(
  name='Infra',
  full_name='feast.core.Infra',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='infra_objects', full_name='feast.core.Infra.infra_objects', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=139,
  serialized_end=194,
)


_INFRAOBJECT_CUSTOMINFRA = _descriptor.Descriptor(
  name='CustomInfra',
  full_name='feast.core.InfraObject.CustomInfra',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='field', full_name='feast.core.InfraObject.CustomInfra.field', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=463,
  serialized_end=491,
)

_INFRAOBJECT = _descriptor.Descriptor(
  name='InfraObject',
  full_name='feast.core.InfraObject',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='infra_object_class_type', full_name='feast.core.InfraObject.infra_object_class_type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dynamodb_table', full_name='feast.core.InfraObject.dynamodb_table', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='datastore_table', full_name='feast.core.InfraObject.datastore_table', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sqlite_table', full_name='feast.core.InfraObject.sqlite_table', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='custom_infra', full_name='feast.core.InfraObject.custom_infra', index=4,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_INFRAOBJECT_CUSTOMINFRA, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='infra_object', full_name='feast.core.InfraObject.infra_object',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=197,
  serialized_end=507,
)

_INFRA.fields_by_name['infra_objects'].message_type = _INFRAOBJECT
_INFRAOBJECT_CUSTOMINFRA.containing_type = _INFRAOBJECT
_INFRAOBJECT.fields_by_name['dynamodb_table'].message_type = feast_dot_core_dot_DynamoDBTable__pb2._DYNAMODBTABLE
_INFRAOBJECT.fields_by_name['datastore_table'].message_type = feast_dot_core_dot_DatastoreTable__pb2._DATASTORETABLE
_INFRAOBJECT.fields_by_name['sqlite_table'].message_type = feast_dot_core_dot_SqliteTable__pb2._SQLITETABLE
_INFRAOBJECT.fields_by_name['custom_infra'].message_type = _INFRAOBJECT_CUSTOMINFRA
_INFRAOBJECT.oneofs_by_name['infra_object'].fields.append(
  _INFRAOBJECT.fields_by_name['dynamodb_table'])
_INFRAOBJECT.fields_by_name['dynamodb_table'].containing_oneof = _INFRAOBJECT.oneofs_by_name['infra_object']
_INFRAOBJECT.oneofs_by_name['infra_object'].fields.append(
  _INFRAOBJECT.fields_by_name['datastore_table'])
_INFRAOBJECT.fields_by_name['datastore_table'].containing_oneof = _INFRAOBJECT.oneofs_by_name['infra_object']
_INFRAOBJECT.oneofs_by_name['infra_object'].fields.append(
  _INFRAOBJECT.fields_by_name['sqlite_table'])
_INFRAOBJECT.fields_by_name['sqlite_table'].containing_oneof = _INFRAOBJECT.oneofs_by_name['infra_object']
_INFRAOBJECT.oneofs_by_name['infra_object'].fields.append(
  _INFRAOBJECT.fields_by_name['custom_infra'])
_INFRAOBJECT.fields_by_name['custom_infra'].containing_oneof = _INFRAOBJECT.oneofs_by_name['infra_object']
DESCRIPTOR.message_types_by_name['Infra'] = _INFRA
DESCRIPTOR.message_types_by_name['InfraObject'] = _INFRAOBJECT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Infra = _reflection.GeneratedProtocolMessageType('Infra', (_message.Message,), {
  'DESCRIPTOR' : _INFRA,
  '__module__' : 'feast.core.InfraObject_pb2'
  # @@protoc_insertion_point(class_scope:feast.core.Infra)
  })
_sym_db.RegisterMessage(Infra)

InfraObject = _reflection.GeneratedProtocolMessageType('InfraObject', (_message.Message,), {

  'CustomInfra' : _reflection.GeneratedProtocolMessageType('CustomInfra', (_message.Message,), {
    'DESCRIPTOR' : _INFRAOBJECT_CUSTOMINFRA,
    '__module__' : 'feast.core.InfraObject_pb2'
    # @@protoc_insertion_point(class_scope:feast.core.InfraObject.CustomInfra)
    })
  ,
  'DESCRIPTOR' : _INFRAOBJECT,
  '__module__' : 'feast.core.InfraObject_pb2'
  # @@protoc_insertion_point(class_scope:feast.core.InfraObject)
  })
_sym_db.RegisterMessage(InfraObject)
_sym_db.RegisterMessage(InfraObject.CustomInfra)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: feast/core/SavedDataset.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from feast.protos.feast.core import DataSource_pb2 as feast_dot_core_dot_DataSource__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='feast/core/SavedDataset.proto',
  package='feast.core',
  syntax='proto3',
  serialized_options=b'\n\020feast.proto.coreB\021SavedDatasetProtoZ3github.com/feast-dev/feast/sdk/go/protos/feast/core',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1d\x66\x65\x61st/core/SavedDataset.proto\x12\nfeast.core\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1b\x66\x65\x61st/core/DataSource.proto\"\xa5\x02\n\x10SavedDatasetSpec\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07project\x18\x02 \x01(\t\x12\x10\n\x08\x66\x65\x61tures\x18\x03 \x03(\t\x12\x11\n\tjoin_keys\x18\x04 \x03(\t\x12\x1a\n\x12\x66ull_feature_names\x18\x05 \x01(\x08\x12\x30\n\x07storage\x18\x06 \x01(\x0b\x32\x1f.feast.core.SavedDatasetStorage\x12\x1c\n\x14\x66\x65\x61ture_service_name\x18\x08 \x01(\t\x12\x34\n\x04tags\x18\x07 \x03(\x0b\x32&.feast.core.SavedDatasetSpec.TagsEntry\x1a+\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xa7\x02\n\x13SavedDatasetStorage\x12:\n\x0c\x66ile_storage\x18\x04 \x01(\x0b\x32\".feast.core.DataSource.FileOptionsH\x00\x12\x42\n\x10\x62igquery_storage\x18\x05 \x01(\x0b\x32&.feast.core.DataSource.BigQueryOptionsH\x00\x12\x42\n\x10redshift_storage\x18\x06 \x01(\x0b\x32&.feast.core.DataSource.RedshiftOptionsH\x00\x12\x44\n\x11snowflake_storage\x18\x07 \x01(\x0b\x32\'.feast.core.DataSource.SnowflakeOptionsH\x00\x42\x06\n\x04kind\"\xf7\x01\n\x10SavedDatasetMeta\x12\x35\n\x11\x63reated_timestamp\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12:\n\x16last_updated_timestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x37\n\x13min_event_timestamp\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x37\n\x13max_event_timestamp\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"f\n\x0cSavedDataset\x12*\n\x04spec\x18\x01 \x01(\x0b\x32\x1c.feast.core.SavedDatasetSpec\x12*\n\x04meta\x18\x02 \x01(\x0b\x32\x1c.feast.core.SavedDatasetMetaBZ\n\x10\x66\x65\x61st.proto.coreB\x11SavedDatasetProtoZ3github.com/feast-dev/feast/sdk/go/protos/feast/coreb\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,feast_dot_core_dot_DataSource__pb2.DESCRIPTOR,])




_SAVEDDATASETSPEC_TAGSENTRY = _descriptor.Descriptor(
  name='TagsEntry',
  full_name='feast.core.SavedDatasetSpec.TagsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='feast.core.SavedDatasetSpec.TagsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='feast.core.SavedDatasetSpec.TagsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=358,
  serialized_end=401,
)

_SAVEDDATASETSPEC = _descriptor.Descriptor(
  name='SavedDatasetSpec',
  full_name='feast.core.SavedDatasetSpec',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='feast.core.SavedDatasetSpec.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='project', full_name='feast.core.SavedDatasetSpec.project', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='features', full_name='feast.core.SavedDatasetSpec.features', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='join_keys', full_name='feast.core.SavedDatasetSpec.join_keys', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='full_feature_names', full_name='feast.core.SavedDatasetSpec.full_feature_names', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='storage', full_name='feast.core.SavedDatasetSpec.storage', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_service_name', full_name='feast.core.SavedDatasetSpec.feature_service_name', index=6,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tags', full_name='feast.core.SavedDatasetSpec.tags', index=7,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_SAVEDDATASETSPEC_TAGSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=108,
  serialized_end=401,
)


_SAVEDDATASETSTORAGE = _descriptor.Descriptor(
  name='SavedDatasetStorage',
  full_name='feast.core.SavedDatasetStorage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='file_storage', full_name='feast.core.SavedDatasetStorage.file_storage', index=0,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bigquery_storage', full_name='feast.core.SavedDatasetStorage.bigquery_storage', index=1,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='redshift_storage', full_name='feast.core.SavedDatasetStorage.redshift_storage', index=2,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='snowflake_storage', full_name='feast.core.SavedDatasetStorage.snowflake_storage', index=3,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='kind', full_name='feast.core.SavedDatasetStorage.kind',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=404,
  serialized_end=699,
)


_SAVEDDATASETMETA = _descriptor.Descriptor(
  name='SavedDatasetMeta',
  full_name='feast.core.SavedDatasetMeta',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='created_timestamp', full_name='feast.core.SavedDatasetMeta.created_timestamp', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_updated_timestamp', full_name='feast.core.SavedDatasetMeta.last_updated_timestamp', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='min_event_timestamp', full_name='feast.core.SavedDatasetMeta.min_event_timestamp', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max_event_timestamp', full_name='feast.core.SavedDatasetMeta.max_event_timestamp', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=702,
  serialized_end=949,
)


_SAVEDDATASET = _descriptor.Descriptor(
  name='SavedDataset',
  full_name='feast.core.SavedDataset',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='spec', full_name='feast.core.SavedDataset.spec', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='meta', full_name='feast.core.SavedDataset.meta', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=951,
  serialized_end=1053,
)

_SAVEDDATASETSPEC_TAGSENTRY.containing_type = _SAVEDDATASETSPEC
_SAVEDDATASETSPEC.fields_by_name['storage'].message_type = _SAVEDDATASETSTORAGE
_SAVEDDATASETSPEC.fields_by_name['tags'].message_type = _SAVEDDATASETSPEC_TAGSENTRY
_SAVEDDATASETSTORAGE.fields_by_name['file_storage'].message_type = feast_dot_core_dot_DataSource__pb2._DATASOURCE_FILEOPTIONS
_SAVEDDATASETSTORAGE.fields_by_name['bigquery_storage'].message_type = feast_dot_core_dot_DataSource__pb2._DATASOURCE_BIGQUERYOPTIONS
_SAVEDDATASETSTORAGE.fields_by_name['redshift_storage'].message_type = feast_dot_core_dot_DataSource__pb2._DATASOURCE_REDSHIFTOPTIONS
_SAVEDDATASETSTORAGE.fields_by_name['snowflake_storage'].message_type = feast_dot_core_dot_DataSource__pb2._DATASOURCE_SNOWFLAKEOPTIONS
_SAVEDDATASETSTORAGE.oneofs_by_name['kind'].fields.append(
  _SAVEDDATASETSTORAGE.fields_by_name['file_storage'])
_SAVEDDATASETSTORAGE.fields_by_name['file_storage'].containing_oneof = _SAVEDDATASETSTORAGE.oneofs_by_name['kind']
_SAVEDDATASETSTORAGE.oneofs_by_name['kind'].fields.append(
  _SAVEDDATASETSTORAGE.fields_by_name['bigquery_storage'])
_SAVEDDATASETSTORAGE.fields_by_name['bigquery_storage'].containing_oneof = _SAVEDDATASETSTORAGE.oneofs_by_name['kind']
_SAVEDDATASETSTORAGE.oneofs_by_name['kind'].fields.append(
  _SAVEDDATASETSTORAGE.fields_by_name['redshift_storage'])
_SAVEDDATASETSTORAGE.fields_by_name['redshift_storage'].containing_oneof = _SAVEDDATASETSTORAGE.oneofs_by_name['kind']
_SAVEDDATASETSTORAGE.oneofs_by_name['kind'].fields.append(
  _SAVEDDATASETSTORAGE.fields_by_name['snowflake_storage'])
_SAVEDDATASETSTORAGE.fields_by_name['snowflake_storage'].containing_oneof = _SAVEDDATASETSTORAGE.oneofs_by_name['kind']
_SAVEDDATASETMETA.fields_by_name['created_timestamp'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_SAVEDDATASETMETA.fields_by_name['last_updated_timestamp'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_SAVEDDATASETMETA.fields_by_name['min_event_timestamp'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_SAVEDDATASETMETA.fields_by_name['max_event_timestamp'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_SAVEDDATASET.fields_by_name['spec'].message_type = _SAVEDDATASETSPEC
_SAVEDDATASET.fields_by_name['meta'].message_type = _SAVEDDATASETMETA
DESCRIPTOR.message_types_by_name['SavedDatasetSpec'] = _SAVEDDATASETSPEC
DESCRIPTOR.message_types_by_name['SavedDatasetStorage'] = _SAVEDDATASETSTORAGE
DESCRIPTOR.message_types_by_name['SavedDatasetMeta'] = _SAVEDDATASETMETA
DESCRIPTOR.message_types_by_name['SavedDataset'] = _SAVEDDATASET
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

SavedDatasetSpec = _reflection.GeneratedProtocolMessageType('SavedDatasetSpec', (_message.Message,), {

  'TagsEntry' : _reflection.GeneratedProtocolMessageType('TagsEntry', (_message.Message,), {
    'DESCRIPTOR' : _SAVEDDATASETSPEC_TAGSENTRY,
    '__module__' : 'feast.core.SavedDataset_pb2'
    # @@protoc_insertion_point(class_scope:feast.core.SavedDatasetSpec.TagsEntry)
    })
  ,
  'DESCRIPTOR' : _SAVEDDATASETSPEC,
  '__module__' : 'feast.core.SavedDataset_pb2'
  # @@protoc_insertion_point(class_scope:feast.core.SavedDatasetSpec)
  })
_sym_db.RegisterMessage(SavedDatasetSpec)
_sym_db.RegisterMessage(SavedDatasetSpec.TagsEntry)

SavedDatasetStorage = _reflection.GeneratedProtocolMessageType('SavedDatasetStorage', (_message.Message,), {
  'DESCRIPTOR' : _SAVEDDATASETSTORAGE,
  '__module__' : 'feast.core.SavedDataset_pb2'
  # @@protoc_insertion_point(class_scope:feast.core.SavedDatasetStorage)
  })
_sym_db.RegisterMessage(SavedDatasetStorage)

SavedDatasetMeta = _reflection.GeneratedProtocolMessageType('SavedDatasetMeta', (_message.Message,), {
  'DESCRIPTOR' : _SAVEDDATASETMETA,
  '__module__' : 'feast.core.SavedDataset_pb2'
  # @@protoc_insertion_point(class_scope:feast.core.SavedDatasetMeta)
  })
_sym_db.RegisterMessage(SavedDatasetMeta)

SavedDataset = _reflection.GeneratedProtocolMessageType('SavedDataset', (_message.Message,), {
  'DESCRIPTOR' : _SAVEDDATASET,
  '__module__' : 'feast.core.SavedDataset_pb2'
  # @@protoc_insertion_point(class_scope:feast.core.SavedDataset)
  })
_sym_db.RegisterMessage(SavedDataset)


DESCRIPTOR._options = None
_SAVEDDATASETSPEC_TAGSENTRY._options = None
# @@protoc_insertion_point(module_scope)



from abc import abstractmethod
from datetime import datetime
from typing import TYPE_CHECKING, Dict, List, Optional, Type, cast

import pandas as pd
import pyarrow
from google.protobuf.json_format import MessageToJson

from feast.data_source import DataSource
from feast.dqm.profilers.profiler import Profile, Profiler
from feast.protos.feast.core.SavedDataset_pb2 import SavedDataset as SavedDatasetProto
from feast.protos.feast.core.SavedDataset_pb2 import SavedDatasetMeta, SavedDatasetSpec
from feast.protos.feast.core.SavedDataset_pb2 import (
    SavedDatasetStorage as SavedDatasetStorageProto,
)

if TYPE_CHECKING:
    from feast.infra.offline_stores.offline_store import RetrievalJob


class _StorageRegistry(type):
    classes_by_proto_attr_name: Dict[str, Type["SavedDatasetStorage"]] = {}

    def __new__(cls, name, bases, dct):
        kls = type.__new__(cls, name, bases, dct)
        if dct.get("_proto_attr_name"):
            cls.classes_by_proto_attr_name[dct["_proto_attr_name"]] = kls
        return kls


class SavedDatasetStorage(metaclass=_StorageRegistry):
    _proto_attr_name: str

    @staticmethod
    def from_proto(storage_proto: SavedDatasetStorageProto) -> "SavedDatasetStorage":
        proto_attr_name = cast(str, storage_proto.WhichOneof("kind"))
        return _StorageRegistry.classes_by_proto_attr_name[proto_attr_name].from_proto(
            storage_proto
        )

    @abstractmethod
    def to_proto(self) -> SavedDatasetStorageProto:
        ...

    @abstractmethod
    def to_data_source(self) -> DataSource:
        ...


class SavedDataset:
    name: str
    features: List[str]
    join_keys: List[str]
    full_feature_names: bool
    storage: SavedDatasetStorage
    tags: Dict[str, str]
    feature_service_name: Optional[str] = None

    created_timestamp: Optional[datetime] = None
    last_updated_timestamp: Optional[datetime] = None

    min_event_timestamp: Optional[datetime] = None
    max_event_timestamp: Optional[datetime] = None

    _retrieval_job: Optional["RetrievalJob"] = None

    def __init__(
        self,
        name: str,
        features: List[str],
        join_keys: List[str],
        storage: SavedDatasetStorage,
        full_feature_names: bool = False,
        tags: Optional[Dict[str, str]] = None,
        feature_service_name: Optional[str] = None,
    ):
        self.name = name
        self.features = features
        self.join_keys = join_keys
        self.storage = storage
        self.full_feature_names = full_feature_names
        self.tags = tags or {}
        self.feature_service_name = feature_service_name

        self._retrieval_job = None

    def __repr__(self):
        items = (f"{k} = {v}" for k, v in self.__dict__.items())
        return f"<{self.__class__.__name__}({', '.join(items)})>"

    def __str__(self):
        return str(MessageToJson(self.to_proto()))

    def __hash__(self):
        return hash((id(self), self.name))

    def __eq__(self, other):
        if not isinstance(other, SavedDataset):
            raise TypeError(
                "Comparisons should only involve FeatureService class objects."
            )
        if self.name != other.name:
            return False

        if sorted(self.features) != sorted(other.features):
            return False

        return True

    @staticmethod
    def from_proto(saved_dataset_proto: SavedDatasetProto):
        """
        Converts a SavedDatasetProto to a SavedDataset object.

        Args:
            saved_dataset_proto: A protobuf representation of a SavedDataset.
        """
        ds = SavedDataset(
            name=saved_dataset_proto.spec.name,
            features=list(saved_dataset_proto.spec.features),
            join_keys=list(saved_dataset_proto.spec.join_keys),
            full_feature_names=saved_dataset_proto.spec.full_feature_names,
            storage=SavedDatasetStorage.from_proto(saved_dataset_proto.spec.storage),
            tags=dict(saved_dataset_proto.spec.tags.items()),
        )

        if saved_dataset_proto.spec.feature_service_name:
            ds.feature_service_name = saved_dataset_proto.spec.feature_service_name

        if saved_dataset_proto.meta.HasField("created_timestamp"):
            ds.created_timestamp = (
                saved_dataset_proto.meta.created_timestamp.ToDatetime()
            )
        if saved_dataset_proto.meta.HasField("last_updated_timestamp"):
            ds.last_updated_timestamp = (
                saved_dataset_proto.meta.last_updated_timestamp.ToDatetime()
            )
        if saved_dataset_proto.meta.HasField("min_event_timestamp"):
            ds.min_event_timestamp = (
                saved_dataset_proto.meta.min_event_timestamp.ToDatetime()
            )
        if saved_dataset_proto.meta.HasField("max_event_timestamp"):
            ds.max_event_timestamp = (
                saved_dataset_proto.meta.max_event_timestamp.ToDatetime()
            )

        return ds

    def to_proto(self) -> SavedDatasetProto:
        """
        Converts a SavedDataset to its protobuf representation.

        Returns:
            A SavedDatasetProto protobuf.
        """
        meta = SavedDatasetMeta()
        if self.created_timestamp:
            meta.created_timestamp.FromDatetime(self.created_timestamp)
        if self.min_event_timestamp:
            meta.min_event_timestamp.FromDatetime(self.min_event_timestamp)
        if self.max_event_timestamp:
            meta.max_event_timestamp.FromDatetime(self.max_event_timestamp)

        spec = SavedDatasetSpec(
            name=self.name,
            features=self.features,
            join_keys=self.join_keys,
            full_feature_names=self.full_feature_names,
            storage=self.storage.to_proto(),
            tags=self.tags,
        )
        if self.feature_service_name:
            spec.feature_service_name = self.feature_service_name

        feature_service_proto = SavedDatasetProto(spec=spec, meta=meta)
        return feature_service_proto

    def with_retrieval_job(self, retrieval_job: "RetrievalJob") -> "SavedDataset":
        self._retrieval_job = retrieval_job
        return self

    def to_df(self) -> pd.DataFrame:
        if not self._retrieval_job:
            raise RuntimeError(
                "To load this dataset use FeatureStore.get_saved_dataset() "
                "instead of instantiating it directly."
            )

        return self._retrieval_job.to_df()

    def to_arrow(self) -> pyarrow.Table:
        if not self._retrieval_job:
            raise RuntimeError(
                "To load this dataset use FeatureStore.get_saved_dataset() "
                "instead of instantiating it directly."
            )

        return self._retrieval_job.to_arrow()

    def as_reference(self, profiler: "Profiler") -> "ValidationReference":
        return ValidationReference(profiler=profiler, dataset=self)

    def get_profile(self, profiler: Profiler) -> Profile:
        return profiler.analyze_dataset(self.to_df())


class ValidationReference:
    dataset: SavedDataset
    profiler: Profiler

    def __init__(self, dataset: SavedDataset, profiler: Profiler):
        self.dataset = dataset
        self.profiler = profiler

    @property
    def profile(self) -> Profile:
        return self.profiler.analyze_dataset(self.dataset.to_df())



import pandas as pd
import pytest
from great_expectations.core import ExpectationSuite
from great_expectations.dataset import PandasDataset

from feast.dqm.errors import ValidationFailed
from feast.dqm.profilers.ge_profiler import ge_profiler
from tests.integration.feature_repos.repo_configuration import (
    construct_universal_feature_views,
)
from tests.integration.feature_repos.universal.entities import (
    customer,
    driver,
    location,
)

_features = [
    "customer_profile:current_balance",
    "customer_profile:avg_passenger_count",
    "customer_profile:lifetime_trip_count",
    "order:order_is_success",
    "global_stats:num_rides",
    "global_stats:avg_ride_length",
]


@ge_profiler
def configurable_profiler(dataset: PandasDataset) -> ExpectationSuite:
    from great_expectations.profile.user_configurable_profiler import (
        UserConfigurableProfiler,
    )

    return UserConfigurableProfiler(
        profile_dataset=dataset,
        excluded_expectations=[
            "expect_table_columns_to_match_ordered_list",
            "expect_table_row_count_to_be_between",
        ],
        value_set_threshold="few",
    ).build_suite()


@ge_profiler
def profiler_with_unrealistic_expectations(dataset: PandasDataset) -> ExpectationSuite:
    # need to create dataframe with corrupted data first
    df = pd.DataFrame()
    df["current_balance"] = [-100]
    df["avg_passenger_count"] = [0]

    other_ds = PandasDataset(df)
    other_ds.expect_column_max_to_be_between("current_balance", -1000, -100)
    other_ds.expect_column_values_to_be_in_set("avg_passenger_count", value_set={0})

    # this should pass
    other_ds.expect_column_min_to_be_between("avg_passenger_count", 0, 1000)

    return other_ds.get_expectation_suite()


@pytest.mark.integration
@pytest.mark.universal
def test_historical_retrieval_with_validation(environment, universal_data_sources):
    store = environment.feature_store

    (entities, datasets, data_sources) = universal_data_sources
    feature_views = construct_universal_feature_views(data_sources)

    store.apply([driver(), customer(), location(), *feature_views.values()])

    entity_df = datasets.entity_df.drop(
        columns=["order_id", "origin_id", "destination_id"]
    )

    reference_job = store.get_historical_features(
        entity_df=entity_df, features=_features,
    )

    store.create_saved_dataset(
        from_=reference_job,
        name="my_training_dataset",
        storage=environment.data_source_creator.create_saved_dataset_destination(),
    )

    job = store.get_historical_features(entity_df=entity_df, features=_features,)

    # if validation pass there will be no exceptions on this point
    job.to_df(
        validation_reference=store.get_saved_dataset(
            "my_training_dataset"
        ).as_reference(profiler=configurable_profiler)
    )


@pytest.mark.integration
@pytest.mark.universal
def test_historical_retrieval_fails_on_validation(environment, universal_data_sources):
    store = environment.feature_store

    (entities, datasets, data_sources) = universal_data_sources
    feature_views = construct_universal_feature_views(data_sources)

    store.apply([driver(), customer(), location(), *feature_views.values()])

    entity_df = datasets.entity_df.drop(
        columns=["order_id", "origin_id", "destination_id"]
    )

    reference_job = store.get_historical_features(
        entity_df=entity_df, features=_features,
    )

    store.create_saved_dataset(
        from_=reference_job,
        name="my_other_dataset",
        storage=environment.data_source_creator.create_saved_dataset_destination(),
    )

    job = store.get_historical_features(entity_df=entity_df, features=_features,)

    with pytest.raises(ValidationFailed) as exc_info:
        job.to_df(
            validation_reference=store.get_saved_dataset(
                "my_other_dataset"
            ).as_reference(profiler=profiler_with_unrealistic_expectations)
        )

    failed_expectations = exc_info.value.report.errors
    assert len(failed_expectations) == 2

    assert failed_expectations[0].check_name == "expect_column_max_to_be_between"
    assert failed_expectations[0].column_name == "current_balance"

    assert failed_expectations[1].check_name == "expect_column_values_to_be_in_set"
    assert failed_expectations[1].column_name == "avg_passenger_count"



import logging
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Any, Dict, List, Tuple, Union

import numpy as np
import pandas as pd
import pyarrow as pa
import pytest

from feast.infra.offline_stores.offline_store import RetrievalJob
from feast.value_type import ValueType
from tests.data.data_creator import create_dataset
from tests.integration.feature_repos.repo_configuration import (
    FULL_REPO_CONFIGS,
    REDIS_CONFIG,
    IntegrationTestRepoConfig,
    construct_test_environment,
)
from tests.integration.feature_repos.universal.entities import driver
from tests.integration.feature_repos.universal.feature_views import driver_feature_view

logger = logging.getLogger(__name__)


def populate_test_configs(offline: bool):
    entity_type_feature_dtypes = [
        (ValueType.INT32, "int32"),
        (ValueType.INT64, "int64"),
        (ValueType.STRING, "float"),
        (ValueType.STRING, "bool"),
        (ValueType.INT32, "datetime"),
    ]
    configs: List[TypeTestConfig] = []
    for test_repo_config in FULL_REPO_CONFIGS:
        for entity_type, feature_dtype in entity_type_feature_dtypes:
            for feature_is_list in [True, False]:
                # Redshift doesn't support list features
                if test_repo_config.provider == "aws" and feature_is_list is True:
                    continue
                # For offline tests, don't need to vary for online store
                if offline and test_repo_config.online_store == REDIS_CONFIG:
                    continue
                for has_empty_list in [True, False]:
                    # For non list features `has_empty_list` does nothing
                    if feature_is_list is False and has_empty_list is True:
                        continue
                    configs.append(
                        TypeTestConfig(
                            entity_type=entity_type,
                            feature_dtype=feature_dtype,
                            feature_is_list=feature_is_list,
                            has_empty_list=has_empty_list,
                            test_repo_config=test_repo_config,
                        )
                    )
    return configs


@dataclass(frozen=True, repr=True)
class TypeTestConfig:
    entity_type: ValueType
    feature_dtype: str
    feature_is_list: bool
    has_empty_list: bool
    test_repo_config: IntegrationTestRepoConfig


OFFLINE_TYPE_TEST_CONFIGS: List[TypeTestConfig] = populate_test_configs(offline=True)
ONLINE_TYPE_TEST_CONFIGS: List[TypeTestConfig] = populate_test_configs(offline=False)


@pytest.fixture(
    params=OFFLINE_TYPE_TEST_CONFIGS,
    scope="session",
    ids=[str(c) for c in OFFLINE_TYPE_TEST_CONFIGS],
)
def offline_types_test_fixtures(request):
    return get_fixtures(request)


@pytest.fixture(
    params=ONLINE_TYPE_TEST_CONFIGS,
    scope="session",
    ids=[str(c) for c in ONLINE_TYPE_TEST_CONFIGS],
)
def online_types_test_fixtures(request):
    return get_fixtures(request)


def get_fixtures(request):
    config: TypeTestConfig = request.param
    # Lower case needed because Redshift lower-cases all table names
    test_project_id = f"{config.entity_type}{config.feature_dtype}{config.feature_is_list}".replace(
        ".", ""
    ).lower()
    type_test_environment = construct_test_environment(
        test_repo_config=config.test_repo_config,
        test_suite_name=f"test_{test_project_id}",
    )
    config = request.param
    df = create_dataset(
        config.entity_type,
        config.feature_dtype,
        config.feature_is_list,
        config.has_empty_list,
    )
    data_source = type_test_environment.data_source_creator.create_data_source(
        df,
        destination_name=type_test_environment.feature_store.project,
        field_mapping={"ts_1": "ts"},
    )
    fv = create_feature_view(
        request.fixturename,
        config.feature_dtype,
        config.feature_is_list,
        config.has_empty_list,
        data_source,
    )

    def cleanup():
        try:
            type_test_environment.data_source_creator.teardown()
        except Exception:  # noqa
            logger.exception("DataSourceCreator teardown has failed")

        type_test_environment.feature_store.teardown()

    request.addfinalizer(cleanup)

    return type_test_environment, config, data_source, fv


@pytest.mark.integration
@pytest.mark.universal
def test_entity_inference_types_match(offline_types_test_fixtures):
    environment, config, data_source, fv = offline_types_test_fixtures
    fs = environment.feature_store

    # Don't specify value type in entity to force inference
    entity = driver(value_type=ValueType.UNKNOWN)
    fs.apply([fv, entity])

    entities = fs.list_entities()
    entity_type_to_expected_inferred_entity_type = {
        ValueType.INT32: ValueType.INT64,
        ValueType.INT64: ValueType.INT64,
        ValueType.FLOAT: ValueType.DOUBLE,
        ValueType.STRING: ValueType.STRING,
    }
    for entity in entities:
        assert (
            entity.value_type
            == entity_type_to_expected_inferred_entity_type[config.entity_type]
        )


@pytest.mark.integration
@pytest.mark.universal
def test_feature_get_historical_features_types_match(offline_types_test_fixtures):
    environment, config, data_source, fv = offline_types_test_fixtures
    fs = environment.feature_store
    fv = create_feature_view(
        "get_historical_features_types_match",
        config.feature_dtype,
        config.feature_is_list,
        config.has_empty_list,
        data_source,
    )
    entity = driver()
    fs.apply([fv, entity])

    features = [f"{fv.name}:value"]
    entity_df = pd.DataFrame()
    entity_df["driver_id"] = (
        ["1", "3"] if config.entity_type == ValueType.STRING else [1, 3]
    )
    now = datetime.utcnow()
    ts = pd.Timestamp(now).round("ms")
    entity_df["ts"] = [
        ts - timedelta(hours=4),
        ts - timedelta(hours=2),
    ]
    historical_features = fs.get_historical_features(
        entity_df=entity_df, features=features,
    )
    # Note: Pandas doesn't play well with nan values in ints. BQ will also coerce to floats if there are NaNs
    historical_features_df = historical_features.to_df()
    print(historical_features_df)

    if config.feature_is_list:
        assert_feature_list_types(
            environment.test_repo_config.provider,
            config.feature_dtype,
            historical_features_df,
        )
    else:
        assert_expected_historical_feature_types(
            config.feature_dtype, historical_features_df
        )
    assert_expected_arrow_types(
        environment.test_repo_config.provider,
        config.feature_dtype,
        config.feature_is_list,
        historical_features,
    )


@pytest.mark.integration
@pytest.mark.universal
def test_feature_get_online_features_types_match(online_types_test_fixtures):
    environment, config, data_source, fv = online_types_test_fixtures
    fv = create_feature_view(
        "get_online_features_types_match",
        config.feature_dtype,
        config.feature_is_list,
        config.has_empty_list,
        data_source,
    )
    fs = environment.feature_store
    features = [fv.name + ":value"]
    entity = driver(value_type=config.entity_type)
    fs.apply([fv, entity])
    fs.materialize(
        environment.start_date,
        environment.end_date
        - timedelta(hours=1)  # throwing out last record to make sure
        # we can successfully infer type even from all empty values
    )

    driver_id_value = "1" if config.entity_type == ValueType.STRING else 1
    online_features = fs.get_online_features(
        features=features, entity_rows=[{"driver": driver_id_value}],
    ).to_dict()

    feature_list_dtype_to_expected_online_response_value_type = {
        "int32": int,
        "int64": int,
        "float": float,
        "string": str,
        "bool": bool,
        "datetime": datetime,
    }
    expected_dtype = feature_list_dtype_to_expected_online_response_value_type[
        config.feature_dtype
    ]

    assert len(online_features["value"]) == 1

    if config.feature_is_list:
        for feature in online_features["value"]:
            assert isinstance(feature, list), "Feature value should be a list"
            assert (
                config.has_empty_list or len(feature) > 0
            ), "List of values should not be empty"
            for element in feature:
                assert isinstance(element, expected_dtype)
    else:
        for feature in online_features["value"]:
            assert isinstance(feature, expected_dtype)


def create_feature_view(
    name, feature_dtype, feature_is_list, has_empty_list, data_source
):
    if feature_is_list is True:
        if feature_dtype == "int32":
            value_type = ValueType.INT32_LIST
        elif feature_dtype == "int64":
            value_type = ValueType.INT64_LIST
        elif feature_dtype == "float":
            value_type = ValueType.FLOAT_LIST
        elif feature_dtype == "bool":
            value_type = ValueType.BOOL_LIST
        elif feature_dtype == "datetime":
            value_type = ValueType.UNIX_TIMESTAMP_LIST
    else:
        if feature_dtype == "int32":
            value_type = ValueType.INT32
        elif feature_dtype == "int64":
            value_type = ValueType.INT64
        elif feature_dtype == "float":
            value_type = ValueType.FLOAT
        elif feature_dtype == "bool":
            value_type = ValueType.BOOL
        elif feature_dtype == "datetime":
            value_type = ValueType.UNIX_TIMESTAMP

    return driver_feature_view(data_source, name=name, value_type=value_type,)


def assert_expected_historical_feature_types(
    feature_dtype: str, historical_features_df: pd.DataFrame
):
    print("Asserting historical feature types")
    feature_dtype_to_expected_historical_feature_dtype = {
        "int32": (pd.api.types.is_integer_dtype,),
        "int64": (pd.api.types.is_int64_dtype,),
        "float": (pd.api.types.is_float_dtype,),
        "string": (pd.api.types.is_string_dtype,),
        "bool": (pd.api.types.is_bool_dtype, pd.api.types.is_object_dtype),
        "datetime": (pd.api.types.is_datetime64_any_dtype,),
    }
    dtype_checkers = feature_dtype_to_expected_historical_feature_dtype[feature_dtype]
    assert any(
        check(historical_features_df.dtypes["value"]) for check in dtype_checkers
    )


def assert_feature_list_types(
    provider: str, feature_dtype: str, historical_features_df: pd.DataFrame
):
    print("Asserting historical feature list types")
    feature_list_dtype_to_expected_historical_feature_list_dtype: Dict[
        str, Union[type, Tuple[Union[type, Tuple[Any, ...]], ...]]
    ] = {
        "int32": (
            int,
            np.int64,
        ),  # Can be `np.int64` if from `np.array` rather that `list`
        "int64": (
            int,
            np.int64,
        ),  # Can be `np.int64` if from `np.array` rather that `list`
        "float": float,
        "string": str,
        "bool": (
            bool,
            np.bool_,
        ),  # Can be `np.bool_` if from `np.array` rather that `list`
        "datetime": np.datetime64,
    }
    expected_dtype = feature_list_dtype_to_expected_historical_feature_list_dtype[
        feature_dtype
    ]
    assert pd.api.types.is_object_dtype(historical_features_df.dtypes["value"])
    for feature in historical_features_df.value:
        assert isinstance(feature, (np.ndarray, list))
        for element in feature:
            assert isinstance(element, expected_dtype)


def assert_expected_arrow_types(
    provider: str,
    feature_dtype: str,
    feature_is_list: bool,
    historical_features: RetrievalJob,
):
    print("Asserting historical feature arrow types")
    historical_features_arrow = historical_features.to_arrow()
    print(historical_features_arrow)
    feature_list_dtype_to_expected_historical_feature_arrow_type = {
        "int32": pa.types.is_int64,
        "int64": pa.types.is_int64,
        "float": pa.types.is_float64,
        "string": pa.types.is_string,
        "bool": pa.types.is_boolean,
        "date": pa.types.is_date,
        "datetime": pa.types.is_timestamp,
    }
    arrow_type_checker = feature_list_dtype_to_expected_historical_feature_arrow_type[
        feature_dtype
    ]
    pa_type = historical_features_arrow.schema.field("value").type

    if feature_is_list:
        assert pa.types.is_list(pa_type)
        assert arrow_type_checker(pa_type.value_type)
    else:
        assert arrow_type_checker(pa_type)



# # # # # # # # # # # # # # # # # # # # # # # #
# This is an example feature definition file  #
# # # # # # # # # # # # # # # # # # # # # # # #

from pathlib import Path

from google.protobuf.duration_pb2 import Duration

from feast import Entity, Feature, FeatureView, ValueType
from feast.infra.offline_stores.contrib.spark_offline_store.spark_source import (
    SparkSource,
)

# Constants related to the generated data sets
CURRENT_DIR = Path(__file__).parent


# Entity definitions
driver = Entity(name="driver_id", value_type=ValueType.INT64, description="driver id",)
customer = Entity(
    name="customer_id", value_type=ValueType.INT64, description="customer id",
)

# Sources
driver_hourly_stats = SparkSource(
    name="driver_hourly_stats",
    path=f"{CURRENT_DIR}/data/driver_hourly_stats.parquet",
    file_format="parquet",
    event_timestamp_column="event_timestamp",
    created_timestamp_column="created",
)
customer_daily_profile = SparkSource(
    name="customer_daily_profile",
    path=f"{CURRENT_DIR}/data/customer_daily_profile.parquet",
    file_format="parquet",
    event_timestamp_column="event_timestamp",
    created_timestamp_column="created",
)

# Feature Views
driver_hourly_stats_view = FeatureView(
    name="driver_hourly_stats",
    entities=["driver_id"],
    ttl=Duration(seconds=86400 * 7),  # one week
    features=[
        Feature(name="conv_rate", dtype=ValueType.FLOAT),
        Feature(name="acc_rate", dtype=ValueType.FLOAT),
        Feature(name="avg_daily_trips", dtype=ValueType.INT64),
    ],
    online=True,
    batch_source=driver_hourly_stats,
    tags={},
)
customer_daily_profile_view = FeatureView(
    name="customer_daily_profile",
    entities=["customer_id"],
    ttl=Duration(seconds=86400 * 7),  # one week
    features=[
        Feature(name="current_balance", dtype=ValueType.FLOAT),
        Feature(name="avg_passenger_count", dtype=ValueType.FLOAT),
        Feature(name="lifetime_trip_count", dtype=ValueType.INT64),
    ],
    online=True,
    batch_source=customer_daily_profile,
    tags={},
)



from google.protobuf import wrappers_pb2 as wrappers

from feast.diff.infra_diff import (
    diff_between,
    diff_infra_protos,
    tag_infra_proto_objects_for_keep_delete_add,
)
from feast.diff.property_diff import TransitionType
from feast.infra.online_stores.datastore import DatastoreTable
from feast.infra.online_stores.dynamodb import DynamoDBTable
from feast.protos.feast.core.InfraObject_pb2 import Infra as InfraProto


def test_tag_infra_proto_objects_for_keep_delete_add():
    to_delete = DynamoDBTable(name="to_delete", region="us-west-2").to_proto()
    to_add = DynamoDBTable(name="to_add", region="us-west-2").to_proto()
    unchanged_table = DynamoDBTable(name="unchanged", region="us-west-2").to_proto()
    pre_changed = DynamoDBTable(name="table", region="us-west-2").to_proto()
    post_changed = DynamoDBTable(name="table", region="us-east-2").to_proto()

    keep, delete, add = tag_infra_proto_objects_for_keep_delete_add(
        [to_delete, unchanged_table, pre_changed],
        [to_add, unchanged_table, post_changed],
    )

    assert len(list(keep)) == 2
    assert unchanged_table in keep
    assert post_changed in keep
    assert to_add not in keep
    assert len(list(delete)) == 1
    assert to_delete in delete
    assert unchanged_table not in delete
    assert pre_changed not in delete
    assert len(list(add)) == 1
    assert to_add in add
    assert unchanged_table not in add
    assert post_changed not in add


def test_diff_between_datastore_tables():
    pre_changed = DatastoreTable(
        project="test", name="table", project_id="pre", namespace="pre"
    ).to_proto()
    post_changed = DatastoreTable(
        project="test", name="table", project_id="post", namespace="post"
    ).to_proto()

    infra_object_diff = diff_between(pre_changed, pre_changed, "datastore table")
    infra_object_property_diffs = infra_object_diff.infra_object_property_diffs
    assert len(infra_object_property_diffs) == 0

    infra_object_diff = diff_between(pre_changed, post_changed, "datastore table")
    infra_object_property_diffs = infra_object_diff.infra_object_property_diffs
    assert len(infra_object_property_diffs) == 2

    assert infra_object_property_diffs[0].property_name == "project_id"
    assert infra_object_property_diffs[0].val_existing == wrappers.StringValue(
        value="pre"
    )
    assert infra_object_property_diffs[0].val_declared == wrappers.StringValue(
        value="post"
    )
    assert infra_object_property_diffs[1].property_name == "namespace"
    assert infra_object_property_diffs[1].val_existing == wrappers.StringValue(
        value="pre"
    )
    assert infra_object_property_diffs[1].val_declared == wrappers.StringValue(
        value="post"
    )


def test_diff_infra_protos():
    to_delete = DynamoDBTable(name="to_delete", region="us-west-2")
    to_add = DynamoDBTable(name="to_add", region="us-west-2")
    unchanged_table = DynamoDBTable(name="unchanged", region="us-west-2")
    pre_changed = DatastoreTable(
        project="test", name="table", project_id="pre", namespace="pre"
    )
    post_changed = DatastoreTable(
        project="test", name="table", project_id="post", namespace="post"
    )

    infra_objects_before = [to_delete, unchanged_table, pre_changed]
    infra_objects_after = [to_add, unchanged_table, post_changed]

    infra_proto_before = InfraProto()
    infra_proto_before.infra_objects.extend(
        [obj.to_infra_object_proto() for obj in infra_objects_before]
    )

    infra_proto_after = InfraProto()
    infra_proto_after.infra_objects.extend(
        [obj.to_infra_object_proto() for obj in infra_objects_after]
    )

    infra_diff = diff_infra_protos(infra_proto_before, infra_proto_after)
    infra_object_diffs = infra_diff.infra_object_diffs

    # There should be one addition, one deletion, one unchanged, and one changed.
    assert len(infra_object_diffs) == 4

    additions = [
        infra_object_diff
        for infra_object_diff in infra_object_diffs
        if infra_object_diff.transition_type == TransitionType.CREATE
    ]
    assert len(additions) == 1
    assert not additions[0].current_infra_object
    assert additions[0].new_infra_object == to_add.to_proto()
    assert len(additions[0].infra_object_property_diffs) == 0

    deletions = [
        infra_object_diff
        for infra_object_diff in infra_object_diffs
        if infra_object_diff.transition_type == TransitionType.DELETE
    ]
    assert len(deletions) == 1
    assert deletions[0].current_infra_object == to_delete.to_proto()
    assert not deletions[0].new_infra_object
    assert len(deletions[0].infra_object_property_diffs) == 0

    unchanged = [
        infra_object_diff
        for infra_object_diff in infra_object_diffs
        if infra_object_diff.transition_type == TransitionType.UNCHANGED
    ]
    assert len(unchanged) == 1
    assert unchanged[0].current_infra_object == unchanged_table.to_proto()
    assert unchanged[0].new_infra_object == unchanged_table.to_proto()
    assert len(unchanged[0].infra_object_property_diffs) == 0

    updates = [
        infra_object_diff
        for infra_object_diff in infra_object_diffs
        if infra_object_diff.transition_type == TransitionType.UPDATE
    ]
    assert len(updates) == 1
    assert updates[0].current_infra_object == pre_changed.to_proto()
    assert updates[0].new_infra_object == post_changed.to_proto()
    assert len(updates[0].infra_object_property_diffs) == 2
    assert updates[0].infra_object_property_diffs[0].property_name == "project_id"
    assert updates[0].infra_object_property_diffs[
        0
    ].val_existing == wrappers.StringValue(value="pre")
    assert updates[0].infra_object_property_diffs[
        0
    ].val_declared == wrappers.StringValue(value="post")
    assert updates[0].infra_object_property_diffs[1].property_name == "namespace"
    assert updates[0].infra_object_property_diffs[
        1
    ].val_existing == wrappers.StringValue(value="pre")
    assert updates[0].infra_object_property_diffs[
        1
    ].val_declared == wrappers.StringValue(value="post")



from dataclasses import dataclass
from typing import Dict, List

from feast import FeatureStore
from feast.protos.feast.types.Value_pb2 import Value


@dataclass
class MockFeatureViewProjection:
    join_key_map: Dict[str, str]


@dataclass
class MockFeatureView:
    name: str
    entities: List[str]
    projection: MockFeatureViewProjection


def test__get_unique_entities():
    entity_values = {
        "entity_1": [Value(int64_val=1), Value(int64_val=2), Value(int64_val=1)],
        "entity_2": [
            Value(string_val="1"),
            Value(string_val="2"),
            Value(string_val="1"),
        ],
        "entity_3": [Value(int64_val=8), Value(int64_val=9), Value(int64_val=10)],
    }

    entity_name_to_join_key_map = {"entity_1": "entity_1", "entity_2": "entity_2"}

    fv = MockFeatureView(
        name="fv_1",
        entities=["entity_1", "entity_2"],
        projection=MockFeatureViewProjection(join_key_map={}),
    )

    unique_entities, indexes = FeatureStore._get_unique_entities(
        FeatureStore,
        table=fv,
        join_key_values=entity_values,
        entity_name_to_join_key_map=entity_name_to_join_key_map,
    )

    assert unique_entities == (
        {"entity_1": Value(int64_val=1), "entity_2": Value(string_val="1")},
        {"entity_1": Value(int64_val=2), "entity_2": Value(string_val="2")},
    )
    assert indexes == ([0, 2], [1])



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: feast/core/DatastoreTable.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='feast/core/DatastoreTable.proto',
  package='feast.core',
  syntax='proto3',
  serialized_options=b'\n\020feast.proto.coreB\023DatastoreTableProtoZ3github.com/feast-dev/feast/sdk/go/protos/feast/core',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1f\x66\x65\x61st/core/DatastoreTable.proto\x12\nfeast.core\x1a\x1egoogle/protobuf/wrappers.proto\"\x92\x01\n\x0e\x44\x61tastoreTable\x12\x0f\n\x07project\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x30\n\nproject_id\x18\x03 \x01(\x0b\x32\x1c.google.protobuf.StringValue\x12/\n\tnamespace\x18\x04 \x01(\x0b\x32\x1c.google.protobuf.StringValueB\\\n\x10\x66\x65\x61st.proto.coreB\x13\x44\x61tastoreTableProtoZ3github.com/feast-dev/feast/sdk/go/protos/feast/coreb\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_wrappers__pb2.DESCRIPTOR,])




_DATASTORETABLE = _descriptor.Descriptor(
  name='DatastoreTable',
  full_name='feast.core.DatastoreTable',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='project', full_name='feast.core.DatastoreTable.project', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='feast.core.DatastoreTable.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='project_id', full_name='feast.core.DatastoreTable.project_id', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='namespace', full_name='feast.core.DatastoreTable.namespace', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=80,
  serialized_end=226,
)

_DATASTORETABLE.fields_by_name['project_id'].message_type = google_dot_protobuf_dot_wrappers__pb2._STRINGVALUE
_DATASTORETABLE.fields_by_name['namespace'].message_type = google_dot_protobuf_dot_wrappers__pb2._STRINGVALUE
DESCRIPTOR.message_types_by_name['DatastoreTable'] = _DATASTORETABLE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DatastoreTable = _reflection.GeneratedProtocolMessageType('DatastoreTable', (_message.Message,), {
  'DESCRIPTOR' : _DATASTORETABLE,
  '__module__' : 'feast.core.DatastoreTable_pb2'
  # @@protoc_insertion_point(class_scope:feast.core.DatastoreTable)
  })
_sym_db.RegisterMessage(DatastoreTable)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# Copyright 2020 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List

import pandas as pd

from feast.feature_view import DUMMY_ENTITY_ID
from feast.protos.feast.serving.ServingService_pb2 import GetOnlineFeaturesResponse
from feast.type_map import feast_value_type_to_python_type

TIMESTAMP_POSTFIX: str = "__ts"


class OnlineResponse:
    """
    Defines an online response in feast.
    """

    def __init__(self, online_response_proto: GetOnlineFeaturesResponse):
        """
        Construct a native online response from its protobuf version.

        Args:
        online_response_proto: GetOnlineResponse proto object to construct from.
        """
        self.proto = online_response_proto
        # Delete DUMMY_ENTITY_ID from proto if it exists
        for idx, val in enumerate(self.proto.metadata.feature_names.val):
            if val == DUMMY_ENTITY_ID:
                del self.proto.metadata.feature_names.val[idx]
                for result in self.proto.results:
                    del result.values[idx]
                    del result.statuses[idx]
                    del result.event_timestamps[idx]
                break

    def to_dict(self, include_event_timestamps: bool = False) -> Dict[str, Any]:
        """
        Converts GetOnlineFeaturesResponse features into a dictionary form.

        Args:
        is_with_event_timestamps: bool Optionally include feature timestamps in the dictionary
        """
        response: Dict[str, List[Any]] = {}

        for result in self.proto.results:
            for idx, feature_ref in enumerate(self.proto.metadata.feature_names.val):
                native_type_value = feast_value_type_to_python_type(result.values[idx])
                if feature_ref not in response:
                    response[feature_ref] = [native_type_value]
                else:
                    response[feature_ref].append(native_type_value)

                if include_event_timestamps:
                    event_ts = result.event_timestamps[idx].seconds
                    timestamp_ref = feature_ref + TIMESTAMP_POSTFIX
                    if timestamp_ref not in response:
                        response[timestamp_ref] = [event_ts]
                    else:
                        response[timestamp_ref].append(event_ts)

        return response

    def to_df(self, include_event_timestamps: bool = False) -> pd.DataFrame:
        """
        Converts GetOnlineFeaturesResponse features into Panda dataframe form.

        Args:
        is_with_event_timestamps: bool Optionally include feature timestamps in the dataframe
        """

        return pd.DataFrame(self.to_dict(include_event_timestamps))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: feast/core/ValidationProfile.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from feast.protos.feast.core import SavedDataset_pb2 as feast_dot_core_dot_SavedDataset__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='feast/core/ValidationProfile.proto',
  package='feast.core',
  syntax='proto3',
  serialized_options=b'\n\020feast.proto.coreB\021ValidationProfileZ3github.com/feast-dev/feast/sdk/go/protos/feast/core',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\"feast/core/ValidationProfile.proto\x12\nfeast.core\x1a\x1d\x66\x65\x61st/core/SavedDataset.proto\"\x83\x01\n\x14GEValidationProfiler\x12\x46\n\x08profiler\x18\x01 \x01(\x0b\x32\x34.feast.core.GEValidationProfiler.UserDefinedProfiler\x1a#\n\x13UserDefinedProfiler\x12\x0c\n\x04\x62ody\x18\x01 \x01(\x0c\"0\n\x13GEValidationProfile\x12\x19\n\x11\x65xpectation_suite\x18\x01 \x01(\x0c\"\x85\x01\n\x13ValidationReference\x12)\n\x07\x64\x61taset\x18\x01 \x01(\x0b\x32\x18.feast.core.SavedDataset\x12\x37\n\x0bge_profiler\x18\x02 \x01(\x0b\x32 .feast.core.GEValidationProfilerH\x00\x42\n\n\x08profilerBZ\n\x10\x66\x65\x61st.proto.coreB\x11ValidationProfileZ3github.com/feast-dev/feast/sdk/go/protos/feast/coreb\x06proto3'
  ,
  dependencies=[feast_dot_core_dot_SavedDataset__pb2.DESCRIPTOR,])




_GEVALIDATIONPROFILER_USERDEFINEDPROFILER = _descriptor.Descriptor(
  name='UserDefinedProfiler',
  full_name='feast.core.GEValidationProfiler.UserDefinedProfiler',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='body', full_name='feast.core.GEValidationProfiler.UserDefinedProfiler.body', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=178,
  serialized_end=213,
)

_GEVALIDATIONPROFILER = _descriptor.Descriptor(
  name='GEValidationProfiler',
  full_name='feast.core.GEValidationProfiler',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='profiler', full_name='feast.core.GEValidationProfiler.profiler', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_GEVALIDATIONPROFILER_USERDEFINEDPROFILER, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=82,
  serialized_end=213,
)


_GEVALIDATIONPROFILE = _descriptor.Descriptor(
  name='GEValidationProfile',
  full_name='feast.core.GEValidationProfile',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='expectation_suite', full_name='feast.core.GEValidationProfile.expectation_suite', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=215,
  serialized_end=263,
)


_VALIDATIONREFERENCE = _descriptor.Descriptor(
  name='ValidationReference',
  full_name='feast.core.ValidationReference',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='dataset', full_name='feast.core.ValidationReference.dataset', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ge_profiler', full_name='feast.core.ValidationReference.ge_profiler', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='profiler', full_name='feast.core.ValidationReference.profiler',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=266,
  serialized_end=399,
)

_GEVALIDATIONPROFILER_USERDEFINEDPROFILER.containing_type = _GEVALIDATIONPROFILER
_GEVALIDATIONPROFILER.fields_by_name['profiler'].message_type = _GEVALIDATIONPROFILER_USERDEFINEDPROFILER
_VALIDATIONREFERENCE.fields_by_name['dataset'].message_type = feast_dot_core_dot_SavedDataset__pb2._SAVEDDATASET
_VALIDATIONREFERENCE.fields_by_name['ge_profiler'].message_type = _GEVALIDATIONPROFILER
_VALIDATIONREFERENCE.oneofs_by_name['profiler'].fields.append(
  _VALIDATIONREFERENCE.fields_by_name['ge_profiler'])
_VALIDATIONREFERENCE.fields_by_name['ge_profiler'].containing_oneof = _VALIDATIONREFERENCE.oneofs_by_name['profiler']
DESCRIPTOR.message_types_by_name['GEValidationProfiler'] = _GEVALIDATIONPROFILER
DESCRIPTOR.message_types_by_name['GEValidationProfile'] = _GEVALIDATIONPROFILE
DESCRIPTOR.message_types_by_name['ValidationReference'] = _VALIDATIONREFERENCE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

GEValidationProfiler = _reflection.GeneratedProtocolMessageType('GEValidationProfiler', (_message.Message,), {

  'UserDefinedProfiler' : _reflection.GeneratedProtocolMessageType('UserDefinedProfiler', (_message.Message,), {
    'DESCRIPTOR' : _GEVALIDATIONPROFILER_USERDEFINEDPROFILER,
    '__module__' : 'feast.core.ValidationProfile_pb2'
    # @@protoc_insertion_point(class_scope:feast.core.GEValidationProfiler.UserDefinedProfiler)
    })
  ,
  'DESCRIPTOR' : _GEVALIDATIONPROFILER,
  '__module__' : 'feast.core.ValidationProfile_pb2'
  # @@protoc_insertion_point(class_scope:feast.core.GEValidationProfiler)
  })
_sym_db.RegisterMessage(GEValidationProfiler)
_sym_db.RegisterMessage(GEValidationProfiler.UserDefinedProfiler)

GEValidationProfile = _reflection.GeneratedProtocolMessageType('GEValidationProfile', (_message.Message,), {
  'DESCRIPTOR' : _GEVALIDATIONPROFILE,
  '__module__' : 'feast.core.ValidationProfile_pb2'
  # @@protoc_insertion_point(class_scope:feast.core.GEValidationProfile)
  })
_sym_db.RegisterMessage(GEValidationProfile)

ValidationReference = _reflection.GeneratedProtocolMessageType('ValidationReference', (_message.Message,), {
  'DESCRIPTOR' : _VALIDATIONREFERENCE,
  '__module__' : 'feast.core.ValidationProfile_pb2'
  # @@protoc_insertion_point(class_scope:feast.core.ValidationReference)
  })
_sym_db.RegisterMessage(ValidationReference)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



import warnings
from typing import Callable, Dict, Iterable, Optional, Tuple

from feast import type_map
from feast.data_source import DataSource
from feast.protos.feast.core.DataSource_pb2 import DataSource as DataSourceProto
from feast.protos.feast.core.SavedDataset_pb2 import (
    SavedDatasetStorage as SavedDatasetStorageProto,
)
from feast.repo_config import RepoConfig
from feast.saved_dataset import SavedDatasetStorage
from feast.value_type import ValueType


class SnowflakeSource(DataSource):
    def __init__(
        self,
        database: Optional[str] = None,
        schema: Optional[str] = None,
        table: Optional[str] = None,
        query: Optional[str] = None,
        event_timestamp_column: Optional[str] = "",
        created_timestamp_column: Optional[str] = "",
        field_mapping: Optional[Dict[str, str]] = None,
        date_partition_column: Optional[str] = "",
        name: Optional[str] = None,
    ):
        """
        Creates a SnowflakeSource object.

        Args:
            database (optional): Snowflake database where the features are stored.
            schema (optional): Snowflake schema in which the table is located.
            table (optional): Snowflake table where the features are stored.
            event_timestamp_column (optional): Event timestamp column used for point in
                time joins of feature values.
            query (optional): The query to be executed to obtain the features.
            created_timestamp_column (optional): Timestamp column indicating when the
                row was created, used for deduplicating rows.
            field_mapping (optional): A dictionary mapping of column names in this data
                source to column names in a feature table or view.
            date_partition_column (optional): Timestamp column used for partitioning.
            name (optional): Name for the source. Defaults to the table if not specified.
        """
        if table is None and query is None:
            raise ValueError('No "table" argument provided.')
        # The default Snowflake schema is named "PUBLIC".
        _schema = "PUBLIC" if (database and table and not schema) else schema

        self.snowflake_options = SnowflakeOptions(
            database=database, schema=_schema, table=table, query=query
        )

        # If no name, use the table as the default name
        _name = name
        if not _name:
            if table:
                _name = table
            else:
                warnings.warn(
                    (
                        f"Starting in Feast 0.21, Feast will require either a name for a data source (if using query) "
                        f"or `table`: {self.query}"
                    ),
                    DeprecationWarning,
                )

        super().__init__(
            name=_name if _name else "",
            event_timestamp_column=event_timestamp_column,
            created_timestamp_column=created_timestamp_column,
            field_mapping=field_mapping,
            date_partition_column=date_partition_column,
        )

    @staticmethod
    def from_proto(data_source: DataSourceProto):
        """
        Creates a SnowflakeSource from a protobuf representation of a SnowflakeSource.

        Args:
            data_source: A protobuf representation of a SnowflakeSource

        Returns:
            A SnowflakeSource object based on the data_source protobuf.
        """
        return SnowflakeSource(
            field_mapping=dict(data_source.field_mapping),
            database=data_source.snowflake_options.database,
            schema=data_source.snowflake_options.schema,
            table=data_source.snowflake_options.table,
            event_timestamp_column=data_source.event_timestamp_column,
            created_timestamp_column=data_source.created_timestamp_column,
            date_partition_column=data_source.date_partition_column,
            query=data_source.snowflake_options.query,
        )

    # Note: Python requires redefining hash in child classes that override __eq__
    def __hash__(self):
        return super().__hash__()

    def __eq__(self, other):
        if not isinstance(other, SnowflakeSource):
            raise TypeError(
                "Comparisons should only involve SnowflakeSource class objects."
            )

        return (
            self.name == other.name
            and self.snowflake_options.database == other.snowflake_options.database
            and self.snowflake_options.schema == other.snowflake_options.schema
            and self.snowflake_options.table == other.snowflake_options.table
            and self.snowflake_options.query == other.snowflake_options.query
            and self.event_timestamp_column == other.event_timestamp_column
            and self.created_timestamp_column == other.created_timestamp_column
            and self.field_mapping == other.field_mapping
        )

    @property
    def database(self):
        """Returns the database of this snowflake source."""
        return self.snowflake_options.database

    @property
    def schema(self):
        """Returns the schema of this snowflake source."""
        return self.snowflake_options.schema

    @property
    def table(self):
        """Returns the table of this snowflake source."""
        return self.snowflake_options.table

    @property
    def query(self):
        """Returns the snowflake options of this snowflake source."""
        return self.snowflake_options.query

    def to_proto(self) -> DataSourceProto:
        """
        Converts a SnowflakeSource object to its protobuf representation.

        Returns:
            A DataSourceProto object.
        """
        data_source_proto = DataSourceProto(
            type=DataSourceProto.BATCH_SNOWFLAKE,
            field_mapping=self.field_mapping,
            snowflake_options=self.snowflake_options.to_proto(),
        )

        data_source_proto.event_timestamp_column = self.event_timestamp_column
        data_source_proto.created_timestamp_column = self.created_timestamp_column
        data_source_proto.date_partition_column = self.date_partition_column

        return data_source_proto

    def validate(self, config: RepoConfig):
        # As long as the query gets successfully executed, or the table exists,
        # the data source is validated. We don't need the results though.
        self.get_table_column_names_and_types(config)

    def get_table_query_string(self) -> str:
        """Returns a string that can directly be used to reference this table in SQL."""
        if self.database and self.table:
            return f'"{self.database}"."{self.schema}"."{self.table}"'
        elif self.table:
            return f'"{self.table}"'
        else:
            return f"({self.query})"

    @staticmethod
    def source_datatype_to_feast_value_type() -> Callable[[str], ValueType]:
        return type_map.snowflake_python_type_to_feast_value_type

    def get_table_column_names_and_types(
        self, config: RepoConfig
    ) -> Iterable[Tuple[str, str]]:
        """
        Returns a mapping of column names to types for this snowflake source.

        Args:
            config: A RepoConfig describing the feature repo
        """

        from feast.infra.offline_stores.snowflake import SnowflakeOfflineStoreConfig
        from feast.infra.utils.snowflake_utils import (
            execute_snowflake_statement,
            get_snowflake_conn,
        )

        assert isinstance(config.offline_store, SnowflakeOfflineStoreConfig)

        snowflake_conn = get_snowflake_conn(config.offline_store)

        if self.database and self.table:
            query = f'SELECT * FROM "{self.database}"."{self.schema}"."{self.table}" LIMIT 1'
        elif self.table:
            query = f'SELECT * FROM "{self.table}" LIMIT 1'
        else:
            query = f"SELECT * FROM ({self.query}) LIMIT 1"

        result = execute_snowflake_statement(snowflake_conn, query).fetch_pandas_all()

        if not result.empty:
            metadata = result.dtypes.apply(str)
            return list(zip(metadata.index, metadata))
        else:
            raise ValueError("The following source:\n" + query + "\n ... is empty")


class SnowflakeOptions:
    """
    DataSource snowflake options used to source features from snowflake query.
    """

    def __init__(
        self,
        database: Optional[str],
        schema: Optional[str],
        table: Optional[str],
        query: Optional[str],
    ):
        self._database = database
        self._schema = schema
        self._table = table
        self._query = query

    @property
    def query(self):
        """Returns the snowflake SQL query referenced by this source."""
        return self._query

    @query.setter
    def query(self, query):
        """Sets the snowflake SQL query referenced by this source."""
        self._query = query

    @property
    def database(self):
        """Returns the database name of this snowflake table."""
        return self._database

    @database.setter
    def database(self, database):
        """Sets the database ref of this snowflake table."""
        self._database = database

    @property
    def schema(self):
        """Returns the schema name of this snowflake table."""
        return self._schema

    @schema.setter
    def schema(self, schema):
        """Sets the schema of this snowflake table."""
        self._schema = schema

    @property
    def table(self):
        """Returns the table name of this snowflake table."""
        return self._table

    @table.setter
    def table(self, table):
        """Sets the table ref of this snowflake table."""
        self._table = table

    @classmethod
    def from_proto(cls, snowflake_options_proto: DataSourceProto.SnowflakeOptions):
        """
        Creates a SnowflakeOptions from a protobuf representation of a snowflake option.

        Args:
            snowflake_options_proto: A protobuf representation of a DataSource

        Returns:
            A SnowflakeOptions object based on the snowflake_options protobuf.
        """
        snowflake_options = cls(
            database=snowflake_options_proto.database,
            schema=snowflake_options_proto.schema,
            table=snowflake_options_proto.table,
            query=snowflake_options_proto.query,
        )

        return snowflake_options

    def to_proto(self) -> DataSourceProto.SnowflakeOptions:
        """
        Converts an SnowflakeOptionsProto object to its protobuf representation.

        Returns:
            A SnowflakeOptionsProto protobuf.
        """
        snowflake_options_proto = DataSourceProto.SnowflakeOptions(
            database=self.database,
            schema=self.schema,
            table=self.table,
            query=self.query,
        )

        return snowflake_options_proto


class SavedDatasetSnowflakeStorage(SavedDatasetStorage):
    _proto_attr_name = "snowflake_storage"

    snowflake_options: SnowflakeOptions

    def __init__(self, table_ref: str):
        self.snowflake_options = SnowflakeOptions(
            database=None, schema=None, table=table_ref, query=None
        )

    @staticmethod
    def from_proto(storage_proto: SavedDatasetStorageProto) -> SavedDatasetStorage:

        return SavedDatasetSnowflakeStorage(
            table_ref=SnowflakeOptions.from_proto(storage_proto.snowflake_storage).table
        )

    def to_proto(self) -> SavedDatasetStorageProto:
        return SavedDatasetStorageProto(
            snowflake_storage=self.snowflake_options.to_proto()
        )

    def to_data_source(self) -> DataSource:
        return SnowflakeSource(table=self.snowflake_options.table)



from feast.diff.registry_diff import (
    diff_registry_objects,
    tag_objects_for_keep_delete_update_add,
)
from feast.feature_view import FeatureView
from tests.utils.data_source_utils import prep_file_source


def test_tag_objects_for_keep_delete_update_add(simple_dataset_1):
    with prep_file_source(
        df=simple_dataset_1, event_timestamp_column="ts_1"
    ) as file_source:
        to_delete = FeatureView(
            name="to_delete", entities=["id"], batch_source=file_source, ttl=None,
        )
        unchanged_fv = FeatureView(
            name="fv1", entities=["id"], batch_source=file_source, ttl=None,
        )
        pre_changed = FeatureView(
            name="fv2",
            entities=["id"],
            batch_source=file_source,
            ttl=None,
            tags={"when": "before"},
        )
        post_changed = FeatureView(
            name="fv2",
            entities=["id"],
            batch_source=file_source,
            ttl=None,
            tags={"when": "after"},
        )
        to_add = FeatureView(
            name="to_add", entities=["id"], batch_source=file_source, ttl=None,
        )

        keep, delete, update, add = tag_objects_for_keep_delete_update_add(
            [unchanged_fv, pre_changed, to_delete], [unchanged_fv, post_changed, to_add]
        )

        assert len(list(keep)) == 2
        assert unchanged_fv in keep
        assert pre_changed in keep
        assert post_changed not in keep
        assert len(list(delete)) == 1
        assert to_delete in delete
        assert len(list(update)) == 2
        assert unchanged_fv in update
        assert post_changed in update
        assert pre_changed not in update
        assert len(list(add)) == 1
        assert to_add in add


def test_diff_registry_objects_feature_views(simple_dataset_1):
    with prep_file_source(
        df=simple_dataset_1, event_timestamp_column="ts_1"
    ) as file_source:
        pre_changed = FeatureView(
            name="fv2",
            entities=["id"],
            batch_source=file_source,
            ttl=None,
            tags={"when": "before"},
        )
        post_changed = FeatureView(
            name="fv2",
            entities=["id"],
            batch_source=file_source,
            ttl=None,
            tags={"when": "after"},
        )

        feast_object_diffs = diff_registry_objects(
            pre_changed, pre_changed, "feature view"
        )
        assert len(feast_object_diffs.feast_object_property_diffs) == 0

        feast_object_diffs = diff_registry_objects(
            pre_changed, post_changed, "feature view"
        )
        assert len(feast_object_diffs.feast_object_property_diffs) == 1

        assert feast_object_diffs.feast_object_property_diffs[0].property_name == "tags"
        assert feast_object_diffs.feast_object_property_diffs[0].val_existing == {
            "when": "before"
        }
        assert feast_object_diffs.feast_object_property_diffs[0].val_declared == {
            "when": "after"
        }



import traceback

import uvicorn
from fastapi import FastAPI, HTTPException, Request
from fastapi.logger import logger
from fastapi.params import Depends
from google.protobuf.json_format import MessageToDict, Parse

import feast
from feast import proto_json
from feast.protos.feast.serving.ServingService_pb2 import GetOnlineFeaturesRequest


def get_app(store: "feast.FeatureStore"):
    proto_json.patch()

    app = FastAPI()

    async def get_body(request: Request):
        return await request.body()

    @app.post("/get-online-features")
    def get_online_features(body=Depends(get_body)):
        try:
            # Validate and parse the request data into GetOnlineFeaturesRequest Protobuf object
            request_proto = GetOnlineFeaturesRequest()
            Parse(body, request_proto)

            # Initialize parameters for FeatureStore.get_online_features(...) call
            if request_proto.HasField("feature_service"):
                features = store.get_feature_service(
                    request_proto.feature_service, allow_cache=True
                )
            else:
                features = list(request_proto.features.val)

            full_feature_names = request_proto.full_feature_names

            batch_sizes = [len(v.val) for v in request_proto.entities.values()]
            num_entities = batch_sizes[0]
            if any(batch_size != num_entities for batch_size in batch_sizes):
                raise HTTPException(status_code=500, detail="Uneven number of columns")

            response_proto = store._get_online_features(
                features,
                request_proto.entities,
                full_feature_names=full_feature_names,
                native_entity_values=False,
            ).proto

            # Convert the Protobuf object to JSON and return it
            return MessageToDict(  # type: ignore
                response_proto, preserving_proto_field_name=True, float_precision=18
            )
        except Exception as e:
            # Print the original exception on the server side
            logger.exception(traceback.format_exc())
            # Raise HTTPException to return the error message to the client
            raise HTTPException(status_code=500, detail=str(e))

    return app


def start_server(
    store: "feast.FeatureStore", host: str, port: int, no_access_log: bool
):
    app = get_app(store)
    uvicorn.run(app, host=host, port=port, access_log=(not no_access_log))



import warnings
from typing import Callable, Dict, Iterable, List, Optional, Tuple

from feast import type_map
from feast.data_source import DataSource
from feast.errors import DataSourceNotFoundException
from feast.protos.feast.core.DataSource_pb2 import DataSource as DataSourceProto
from feast.protos.feast.core.SavedDataset_pb2 import (
    SavedDatasetStorage as SavedDatasetStorageProto,
)
from feast.repo_config import RepoConfig
from feast.saved_dataset import SavedDatasetStorage
from feast.value_type import ValueType


class BigQuerySource(DataSource):
    def __init__(
        self,
        event_timestamp_column: Optional[str] = "",
        table: Optional[str] = None,
        table_ref: Optional[str] = None,
        created_timestamp_column: Optional[str] = "",
        field_mapping: Optional[Dict[str, str]] = None,
        date_partition_column: Optional[str] = "",
        query: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """Create a BigQuerySource from an existing table or query.

         Args:
             table (optional): The BigQuery table where features can be found.
             table_ref (optional): (Deprecated) The BigQuery table where features can be found.
             event_timestamp_column: Event timestamp column used for point in time joins of feature values.
             created_timestamp_column (optional): Timestamp column when row was created, used for deduplicating rows.
             field_mapping: A dictionary mapping of column names in this data source to feature names in a feature table
                 or view. Only used for feature columns, not entities or timestamp columns.
             date_partition_column (optional): Timestamp column used for partitioning.
             query (optional): SQL query to execute to generate data for this data source.
             name (optional): Name for the source. Defaults to the table_ref if not specified.
         Example:
             >>> from feast import BigQuerySource
             >>> my_bigquery_source = BigQuerySource(table="gcp_project:bq_dataset.bq_table")
         """
        if table is None and table_ref is None and query is None:
            raise ValueError('No "table" or "query" argument provided.')
        if not table and table_ref:
            warnings.warn(
                (
                    "The argument 'table_ref' is being deprecated. Please use 'table' "
                    "instead. Feast 0.20 and onwards will not support the argument 'table_ref'."
                ),
                DeprecationWarning,
            )
            table = table_ref
        self.bigquery_options = BigQueryOptions(table_ref=table, query=query)

        # If no name, use the table_ref as the default name
        _name = name
        if not _name:
            if table:
                _name = table
            elif table_ref:
                _name = table_ref
            else:
                warnings.warn(
                    (
                        f"Starting in Feast 0.21, Feast will require either a name for a data source (if using query) or `table`: {self.query}"
                    ),
                    DeprecationWarning,
                )

        super().__init__(
            name=_name if _name else "",
            event_timestamp_column=event_timestamp_column,
            created_timestamp_column=created_timestamp_column,
            field_mapping=field_mapping,
            date_partition_column=date_partition_column,
        )

    # Note: Python requires redefining hash in child classes that override __eq__
    def __hash__(self):
        return super().__hash__()

    def __eq__(self, other):
        if not isinstance(other, BigQuerySource):
            raise TypeError(
                "Comparisons should only involve BigQuerySource class objects."
            )

        return (
            self.name == other.name
            and self.bigquery_options.table_ref == other.bigquery_options.table_ref
            and self.bigquery_options.query == other.bigquery_options.query
            and self.event_timestamp_column == other.event_timestamp_column
            and self.created_timestamp_column == other.created_timestamp_column
            and self.field_mapping == other.field_mapping
        )

    @property
    def table_ref(self):
        return self.bigquery_options.table_ref

    @property
    def query(self):
        return self.bigquery_options.query

    @staticmethod
    def from_proto(data_source: DataSourceProto):

        assert data_source.HasField("bigquery_options")

        return BigQuerySource(
            name=data_source.name,
            field_mapping=dict(data_source.field_mapping),
            table_ref=data_source.bigquery_options.table_ref,
            event_timestamp_column=data_source.event_timestamp_column,
            created_timestamp_column=data_source.created_timestamp_column,
            date_partition_column=data_source.date_partition_column,
            query=data_source.bigquery_options.query,
        )

    def to_proto(self) -> DataSourceProto:
        data_source_proto = DataSourceProto(
            name=self.name,
            type=DataSourceProto.BATCH_BIGQUERY,
            field_mapping=self.field_mapping,
            bigquery_options=self.bigquery_options.to_proto(),
        )

        data_source_proto.event_timestamp_column = self.event_timestamp_column
        data_source_proto.created_timestamp_column = self.created_timestamp_column
        data_source_proto.date_partition_column = self.date_partition_column

        return data_source_proto

    def validate(self, config: RepoConfig):
        if not self.query:
            from google.api_core.exceptions import NotFound
            from google.cloud import bigquery

            client = bigquery.Client()
            try:
                client.get_table(self.table_ref)
            except NotFound:
                raise DataSourceNotFoundException(self.table_ref)

    def get_table_query_string(self) -> str:
        """Returns a string that can directly be used to reference this table in SQL"""
        if self.table_ref:
            return f"`{self.table_ref}`"
        else:
            return f"({self.query})"

    @staticmethod
    def source_datatype_to_feast_value_type() -> Callable[[str], ValueType]:
        return type_map.bq_to_feast_value_type

    def get_table_column_names_and_types(
        self, config: RepoConfig
    ) -> Iterable[Tuple[str, str]]:
        from google.cloud import bigquery

        client = bigquery.Client()
        if self.table_ref is not None:
            schema = client.get_table(self.table_ref).schema
            if not isinstance(schema[0], bigquery.schema.SchemaField):
                raise TypeError("Could not parse BigQuery table schema.")
        else:
            bq_columns_query = f"SELECT * FROM ({self.query}) LIMIT 1"
            queryRes = client.query(bq_columns_query).result()
            schema = queryRes.schema

        name_type_pairs: List[Tuple[str, str]] = []
        for field in schema:
            bq_type_as_str = field.field_type
            if field.mode == "REPEATED":
                bq_type_as_str = "ARRAY<" + bq_type_as_str + ">"
            name_type_pairs.append((field.name, bq_type_as_str))

        return name_type_pairs


class BigQueryOptions:
    """
    DataSource BigQuery options used to source features from BigQuery query
    """

    def __init__(
        self, table_ref: Optional[str], query: Optional[str],
    ):
        self._table_ref = table_ref
        self._query = query

    @property
    def query(self):
        """
        Returns the BigQuery SQL query referenced by this source
        """
        return self._query

    @query.setter
    def query(self, query):
        """
        Sets the BigQuery SQL query referenced by this source
        """
        self._query = query

    @property
    def table_ref(self):
        """
        Returns the table ref of this BQ table
        """
        return self._table_ref

    @table_ref.setter
    def table_ref(self, table_ref):
        """
        Sets the table ref of this BQ table
        """
        self._table_ref = table_ref

    @classmethod
    def from_proto(cls, bigquery_options_proto: DataSourceProto.BigQueryOptions):
        """
        Creates a BigQueryOptions from a protobuf representation of a BigQuery option

        Args:
            bigquery_options_proto: A protobuf representation of a DataSource

        Returns:
            Returns a BigQueryOptions object based on the bigquery_options protobuf
        """

        bigquery_options = cls(
            table_ref=bigquery_options_proto.table_ref,
            query=bigquery_options_proto.query,
        )

        return bigquery_options

    def to_proto(self) -> DataSourceProto.BigQueryOptions:
        """
        Converts an BigQueryOptionsProto object to its protobuf representation.

        Returns:
            BigQueryOptionsProto protobuf
        """

        bigquery_options_proto = DataSourceProto.BigQueryOptions(
            table_ref=self.table_ref, query=self.query,
        )

        return bigquery_options_proto


class SavedDatasetBigQueryStorage(SavedDatasetStorage):
    _proto_attr_name = "bigquery_storage"

    bigquery_options: BigQueryOptions

    def __init__(self, table_ref: str):
        self.bigquery_options = BigQueryOptions(table_ref=table_ref, query=None)

    @staticmethod
    def from_proto(storage_proto: SavedDatasetStorageProto) -> SavedDatasetStorage:
        return SavedDatasetBigQueryStorage(
            table_ref=BigQueryOptions.from_proto(
                storage_proto.bigquery_storage
            ).table_ref
        )

    def to_proto(self) -> SavedDatasetStorageProto:
        return SavedDatasetStorageProto(
            bigquery_storage=self.bigquery_options.to_proto()
        )

    def to_data_source(self) -> DataSource:
        return BigQuerySource(table_ref=self.bigquery_options.table_ref)



import dataclasses
import importlib
import json
import os
import re
import tempfile
import uuid
from dataclasses import dataclass
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, List, Optional, Tuple, Union

import pandas as pd
import yaml

from feast import FeatureStore, FeatureView, OnDemandFeatureView, driver_test_data
from feast.constants import FULL_REPO_CONFIGS_MODULE_ENV_NAME
from feast.data_source import DataSource
from feast.errors import FeastModuleImportError
from feast.repo_config import RegistryConfig, RepoConfig
from tests.integration.feature_repos.integration_test_repo_config import (
    IntegrationTestRepoConfig,
)
from tests.integration.feature_repos.universal.data_source_creator import (
    DataSourceCreator,
)
from tests.integration.feature_repos.universal.data_sources.bigquery import (
    BigQueryDataSourceCreator,
)
from tests.integration.feature_repos.universal.data_sources.redshift import (
    RedshiftDataSourceCreator,
)
from tests.integration.feature_repos.universal.data_sources.snowflake import (
    SnowflakeDataSourceCreator,
)
from tests.integration.feature_repos.universal.feature_views import (
    conv_rate_plus_100_feature_view,
    create_conv_rate_request_data_source,
    create_customer_daily_profile_feature_view,
    create_driver_age_request_feature_view,
    create_driver_hourly_stats_feature_view,
    create_field_mapping_feature_view,
    create_global_stats_feature_view,
    create_location_stats_feature_view,
    create_order_feature_view,
)

DYNAMO_CONFIG = {"type": "dynamodb", "region": "us-west-2"}
# Port 12345 will chosen as default for redis node configuration because Redis Cluster is started off of nodes 6379 -> 6384. This causes conflicts in cli integration tests so we manually keep them separate.
REDIS_CONFIG = {"type": "redis", "connection_string": "localhost:12345,db=0"}
REDIS_CLUSTER_CONFIG = {
    "type": "redis",
    "redis_type": "redis_cluster",
    # Redis Cluster Port Forwarding is setup in "pr_integration_tests.yaml" under "Setup Redis Cluster".
    "connection_string": "127.0.0.1:6001,127.0.0.1:6002,127.0.0.1:6003",
}

# FULL_REPO_CONFIGS contains the repo configurations (e.g. provider, offline store,
# online store, test data, and more parameters) that most integration tests will test
# against. By default, FULL_REPO_CONFIGS uses the three providers (local, GCP, and AWS)
# with their default offline and online stores; it also tests the providers with the
# Redis online store. It can be overwritten by specifying a Python module through the
# FULL_REPO_CONFIGS_MODULE_ENV_NAME environment variable. In this case, that Python
# module will be imported and FULL_REPO_CONFIGS will be extracted from the file.
DEFAULT_FULL_REPO_CONFIGS: List[IntegrationTestRepoConfig] = [
    # Local configurations
    IntegrationTestRepoConfig(),
    IntegrationTestRepoConfig(python_feature_server=True),
]
if os.getenv("FEAST_IS_LOCAL_TEST", "False") != "True":
    DEFAULT_FULL_REPO_CONFIGS.extend(
        [
            IntegrationTestRepoConfig(online_store=REDIS_CONFIG),
            IntegrationTestRepoConfig(online_store=REDIS_CLUSTER_CONFIG),
            # GCP configurations
            IntegrationTestRepoConfig(
                provider="gcp",
                offline_store_creator=BigQueryDataSourceCreator,
                online_store="datastore",
            ),
            IntegrationTestRepoConfig(
                provider="gcp",
                offline_store_creator=BigQueryDataSourceCreator,
                online_store=REDIS_CONFIG,
            ),
            # AWS configurations
            IntegrationTestRepoConfig(
                provider="aws",
                offline_store_creator=RedshiftDataSourceCreator,
                online_store=DYNAMO_CONFIG,
                python_feature_server=True,
            ),
            IntegrationTestRepoConfig(
                provider="aws",
                offline_store_creator=RedshiftDataSourceCreator,
                online_store=REDIS_CONFIG,
            ),
            # Snowflake configurations
            IntegrationTestRepoConfig(
                provider="aws",  # no list features, no feature server
                offline_store_creator=SnowflakeDataSourceCreator,
                online_store=REDIS_CONFIG,
            ),
        ]
    )
full_repo_configs_module = os.environ.get(FULL_REPO_CONFIGS_MODULE_ENV_NAME)
if full_repo_configs_module is not None:
    try:
        module = importlib.import_module(full_repo_configs_module)
        FULL_REPO_CONFIGS = getattr(module, "FULL_REPO_CONFIGS")
    except Exception as e:
        raise FeastModuleImportError(
            "FULL_REPO_CONFIGS", full_repo_configs_module
        ) from e
else:
    FULL_REPO_CONFIGS = DEFAULT_FULL_REPO_CONFIGS


@dataclass
class UniversalEntities:
    customer_vals: List[Any]
    driver_vals: List[Any]
    location_vals: List[Any]


def construct_universal_entities() -> UniversalEntities:
    return UniversalEntities(
        customer_vals=list(range(1001, 1020)),
        driver_vals=list(range(5001, 5020)),
        location_vals=list(range(1, 50)),
    )


@dataclass
class UniversalDatasets:
    customer_df: pd.DataFrame
    driver_df: pd.DataFrame
    location_df: pd.DataFrame
    orders_df: pd.DataFrame
    global_df: pd.DataFrame
    field_mapping_df: pd.DataFrame
    entity_df: pd.DataFrame


def construct_universal_datasets(
    entities: UniversalEntities, start_time: datetime, end_time: datetime
) -> UniversalDatasets:
    customer_df = driver_test_data.create_customer_daily_profile_df(
        entities.customer_vals, start_time, end_time
    )
    driver_df = driver_test_data.create_driver_hourly_stats_df(
        entities.driver_vals, start_time, end_time
    )
    location_df = driver_test_data.create_location_stats_df(
        entities.location_vals, start_time, end_time
    )
    orders_df = driver_test_data.create_orders_df(
        customers=entities.customer_vals,
        drivers=entities.driver_vals,
        locations=entities.location_vals,
        start_date=start_time,
        end_date=end_time,
        order_count=20,
    )
    global_df = driver_test_data.create_global_daily_stats_df(start_time, end_time)
    field_mapping_df = driver_test_data.create_field_mapping_df(start_time, end_time)
    entity_df = orders_df[
        [
            "customer_id",
            "driver_id",
            "order_id",
            "origin_id",
            "destination_id",
            "event_timestamp",
        ]
    ]

    return UniversalDatasets(
        customer_df=customer_df,
        driver_df=driver_df,
        location_df=location_df,
        orders_df=orders_df,
        global_df=global_df,
        field_mapping_df=field_mapping_df,
        entity_df=entity_df,
    )


@dataclass
class UniversalDataSources:
    customer: DataSource
    driver: DataSource
    location: DataSource
    orders: DataSource
    global_ds: DataSource
    field_mapping: DataSource


def construct_universal_data_sources(
    datasets: UniversalDatasets, data_source_creator: DataSourceCreator
) -> UniversalDataSources:
    customer_ds = data_source_creator.create_data_source(
        datasets.customer_df,
        destination_name="customer_profile",
        event_timestamp_column="event_timestamp",
        created_timestamp_column="created",
    )
    driver_ds = data_source_creator.create_data_source(
        datasets.driver_df,
        destination_name="driver_hourly",
        event_timestamp_column="event_timestamp",
        created_timestamp_column="created",
    )
    location_ds = data_source_creator.create_data_source(
        datasets.location_df,
        destination_name="location_hourly",
        event_timestamp_column="event_timestamp",
        created_timestamp_column="created",
    )
    orders_ds = data_source_creator.create_data_source(
        datasets.orders_df,
        destination_name="orders",
        event_timestamp_column="event_timestamp",
        created_timestamp_column=None,
    )
    global_ds = data_source_creator.create_data_source(
        datasets.global_df,
        destination_name="global",
        event_timestamp_column="event_timestamp",
        created_timestamp_column="created",
    )
    field_mapping_ds = data_source_creator.create_data_source(
        datasets.field_mapping_df,
        destination_name="field_mapping",
        event_timestamp_column="event_timestamp",
        created_timestamp_column="created",
        field_mapping={"column_name": "feature_name"},
    )
    return UniversalDataSources(
        customer=customer_ds,
        driver=driver_ds,
        location=location_ds,
        orders=orders_ds,
        global_ds=global_ds,
        field_mapping=field_mapping_ds,
    )


@dataclass
class UniversalFeatureViews:
    customer: FeatureView
    global_fv: FeatureView
    driver: FeatureView
    driver_odfv: OnDemandFeatureView
    driver_age_request_fv: FeatureView
    order: FeatureView
    location: FeatureView
    field_mapping: FeatureView

    def values(self):
        return dataclasses.asdict(self).values()


def construct_universal_feature_views(
    data_sources: UniversalDataSources, with_odfv: bool = True,
) -> UniversalFeatureViews:
    driver_hourly_stats = create_driver_hourly_stats_feature_view(data_sources.driver)
    return UniversalFeatureViews(
        customer=create_customer_daily_profile_feature_view(data_sources.customer),
        global_fv=create_global_stats_feature_view(data_sources.global_ds),
        driver=driver_hourly_stats,
        driver_odfv=conv_rate_plus_100_feature_view(
            {
                "driver": driver_hourly_stats,
                "input_request": create_conv_rate_request_data_source(),
            }
        )
        if with_odfv
        else None,
        driver_age_request_fv=create_driver_age_request_feature_view(),
        order=create_order_feature_view(data_sources.orders),
        location=create_location_stats_feature_view(data_sources.location),
        field_mapping=create_field_mapping_feature_view(data_sources.field_mapping),
    )


@dataclass
class Environment:
    name: str
    test_repo_config: IntegrationTestRepoConfig
    feature_store: FeatureStore
    data_source_creator: DataSourceCreator
    python_feature_server: bool
    worker_id: str

    def __post_init__(self):
        self.end_date = datetime.utcnow().replace(microsecond=0, second=0, minute=0)
        self.start_date: datetime = self.end_date - timedelta(days=3)

    def get_feature_server_endpoint(self) -> str:
        if self.python_feature_server and self.test_repo_config.provider == "local":
            return f"http://localhost:{self.get_local_server_port()}"
        return self.feature_store.get_feature_server_endpoint()

    def get_local_server_port(self) -> int:
        # Heuristic when running with xdist to extract unique ports for each worker
        parsed_worker_id = re.findall("gw(\\d+)", self.worker_id)
        if len(parsed_worker_id) != 0:
            worker_id_num = int(parsed_worker_id[0])
        else:
            worker_id_num = 0
        return 6566 + worker_id_num


def table_name_from_data_source(ds: DataSource) -> Optional[str]:
    if hasattr(ds, "table_ref"):
        return ds.table_ref  # type: ignore
    elif hasattr(ds, "table"):
        return ds.table  # type: ignore
    return None


def construct_test_environment(
    test_repo_config: IntegrationTestRepoConfig,
    test_suite_name: str = "integration_test",
    worker_id: str = "worker_id",
) -> Environment:
    _uuid = str(uuid.uuid4()).replace("-", "")[:6]

    run_id = os.getenv("GITHUB_RUN_ID", default=None)
    run_id = f"gh_run_{run_id}_{_uuid}" if run_id else _uuid
    run_num = os.getenv("GITHUB_RUN_NUMBER", default=1)

    project = f"{test_suite_name}_{run_id}_{run_num}"

    offline_creator: DataSourceCreator = test_repo_config.offline_store_creator(project)

    offline_store_config = offline_creator.create_offline_store_config()
    online_store = test_repo_config.online_store

    repo_dir_name = tempfile.mkdtemp()

    if test_repo_config.python_feature_server and test_repo_config.provider == "aws":
        from feast.infra.feature_servers.aws_lambda.config import (
            AwsLambdaFeatureServerConfig,
        )

        feature_server = AwsLambdaFeatureServerConfig(
            enabled=True,
            execution_role_name="arn:aws:iam::402087665549:role/lambda_execution_role",
        )

        registry = (
            f"s3://feast-integration-tests/registries/{project}/registry.db"
        )  # type: Union[str, RegistryConfig]
    else:
        # Note: even if it's a local feature server, the repo config does not have this configured
        feature_server = None
        registry = RegistryConfig(
            path=str(Path(repo_dir_name) / "registry.db"), cache_ttl_seconds=1,
        )
    config = RepoConfig(
        registry=registry,
        project=project,
        provider=test_repo_config.provider,
        offline_store=offline_store_config,
        online_store=online_store,
        repo_path=repo_dir_name,
        feature_server=feature_server,
    )

    # Create feature_store.yaml out of the config
    with open(Path(repo_dir_name) / "feature_store.yaml", "w") as f:
        yaml.safe_dump(json.loads(config.json()), f)

    fs = FeatureStore(repo_dir_name)
    # We need to initialize the registry, because if nothing is applied in the test before tearing down
    # the feature store, that will cause the teardown method to blow up.
    fs.registry._initialize_registry()
    environment = Environment(
        name=project,
        test_repo_config=test_repo_config,
        feature_store=fs,
        data_source_creator=offline_creator,
        python_feature_server=test_repo_config.python_feature_server,
        worker_id=worker_id,
    )

    return environment


TestData = Tuple[UniversalEntities, UniversalDatasets, UniversalDataSources]


def construct_universal_test_data(environment: Environment) -> TestData:
    entities = construct_universal_entities()
    datasets = construct_universal_datasets(
        entities, environment.start_date, environment.end_date
    )
    data_sources = construct_universal_data_sources(
        datasets, environment.data_source_creator
    )

    return entities, datasets, data_sources



from feast import FileSource

driver_hourly_stats = FileSource(
    path="driver_stats.parquet",  # this parquet is not real and will not be read
)

driver_hourly_stats_clone = FileSource(
    path="driver_stats.parquet",  # this parquet is not real and will not be read
)



import os
import uuid
from typing import Dict, List, Optional

import pandas as pd

from feast import SnowflakeSource
from feast.data_source import DataSource
from feast.infra.offline_stores.snowflake import SnowflakeOfflineStoreConfig
from feast.infra.offline_stores.snowflake_source import SavedDatasetSnowflakeStorage
from feast.infra.utils.snowflake_utils import get_snowflake_conn, write_pandas
from feast.repo_config import FeastConfigBaseModel
from tests.integration.feature_repos.universal.data_source_creator import (
    DataSourceCreator,
)


class SnowflakeDataSourceCreator(DataSourceCreator):

    tables: List[str] = []

    def __init__(self, project_name: str):
        super().__init__()
        self.project_name = project_name
        self.offline_store_config = SnowflakeOfflineStoreConfig(
            type="snowflake.offline",
            account=os.environ["SNOWFLAKE_CI_DEPLOYMENT"],
            user=os.environ["SNOWFLAKE_CI_USER"],
            password=os.environ["SNOWFLAKE_CI_PASSWORD"],
            role=os.environ["SNOWFLAKE_CI_ROLE"],
            warehouse=os.environ["SNOWFLAKE_CI_WAREHOUSE"],
            database="FEAST",
            schema="OFFLINE",
        )

    def create_data_source(
        self,
        df: pd.DataFrame,
        destination_name: str,
        suffix: Optional[str] = None,
        event_timestamp_column="ts",
        created_timestamp_column="created_ts",
        field_mapping: Dict[str, str] = None,
    ) -> DataSource:

        snowflake_conn = get_snowflake_conn(self.offline_store_config)

        destination_name = self.get_prefixed_table_name(destination_name)

        write_pandas(snowflake_conn, df, destination_name, auto_create_table=True)

        self.tables.append(destination_name)

        return SnowflakeSource(
            table=destination_name,
            event_timestamp_column=event_timestamp_column,
            created_timestamp_column=created_timestamp_column,
            date_partition_column="",
            field_mapping=field_mapping or {"ts_1": "ts"},
        )

    def create_saved_dataset_destination(self) -> SavedDatasetSnowflakeStorage:
        table = self.get_prefixed_table_name(
            f"persisted_ds_{str(uuid.uuid4()).replace('-', '_')}"
        )
        self.tables.append(table)

        return SavedDatasetSnowflakeStorage(table_ref=table)

    def create_offline_store_config(self) -> FeastConfigBaseModel:
        return self.offline_store_config

    def get_prefixed_table_name(self, suffix: str) -> str:
        return f"{self.project_name}_{suffix}"

    def teardown(self):
        snowflake_conn = get_snowflake_conn(self.offline_store_config)

        with snowflake_conn as conn:
            cur = conn.cursor()
            for table in self.tables:
                cur.execute(f'DROP TABLE IF EXISTS "{table}"')



import pandas as pd
import pytest

from feast import (
    BigQuerySource,
    Entity,
    Feature,
    FileSource,
    RedshiftSource,
    RepoConfig,
    SnowflakeSource,
    ValueType,
)
from feast.data_source import RequestDataSource
from feast.errors import (
    DataSourceNoNameException,
    RegistryInferenceFailure,
    SpecifiedFeaturesNotPresentError,
)
from feast.feature_view import FeatureView
from feast.inference import (
    update_data_sources_with_inferred_event_timestamp_col,
    update_entities_with_inferred_types_from_feature_views,
)
from feast.infra.offline_stores.contrib.spark_offline_store.spark_source import (
    SparkSource,
)
from feast.on_demand_feature_view import on_demand_feature_view
from tests.utils.data_source_utils import (
    prep_file_source,
    simple_bq_source_using_query_arg,
    simple_bq_source_using_table_ref_arg,
)


def test_update_entities_with_inferred_types_from_feature_views(
    simple_dataset_1, simple_dataset_2
):
    with prep_file_source(
        df=simple_dataset_1, event_timestamp_column="ts_1"
    ) as file_source, prep_file_source(
        df=simple_dataset_2, event_timestamp_column="ts_1"
    ) as file_source_2:

        fv1 = FeatureView(
            name="fv1", entities=["id"], batch_source=file_source, ttl=None,
        )
        fv2 = FeatureView(
            name="fv2", entities=["id"], batch_source=file_source_2, ttl=None,
        )

        actual_1 = Entity(name="id", join_key="id_join_key")
        actual_2 = Entity(name="id", join_key="id_join_key")

        update_entities_with_inferred_types_from_feature_views(
            [actual_1], [fv1], RepoConfig(provider="local", project="test")
        )
        update_entities_with_inferred_types_from_feature_views(
            [actual_2], [fv2], RepoConfig(provider="local", project="test")
        )
        assert actual_1 == Entity(
            name="id", join_key="id_join_key", value_type=ValueType.INT64
        )
        assert actual_2 == Entity(
            name="id", join_key="id_join_key", value_type=ValueType.STRING
        )

        with pytest.raises(RegistryInferenceFailure):
            # two viable data types
            update_entities_with_inferred_types_from_feature_views(
                [Entity(name="id", join_key="id_join_key")],
                [fv1, fv2],
                RepoConfig(provider="local", project="test"),
            )


def test_infer_datasource_names_file():
    file_path = "path/to/test.csv"
    data_source = FileSource(path=file_path)
    assert data_source.name == file_path

    source_name = "my_name"
    data_source = FileSource(name=source_name, path=file_path)
    assert data_source.name == source_name


def test_infer_datasource_names_dwh():
    table = "project.table"
    dwh_classes = [BigQuerySource, RedshiftSource, SnowflakeSource, SparkSource]

    for dwh_class in dwh_classes:
        data_source = dwh_class(table=table)
        assert data_source.name == table

        source_name = "my_name"
        data_source_with_table = dwh_class(name=source_name, table=table)
        assert data_source_with_table.name == source_name
        data_source_with_query = dwh_class(
            name=source_name, query=f"SELECT * from {table}"
        )
        assert data_source_with_query.name == source_name

        # If we have a query and no name, throw an error
        if dwh_class == SparkSource:
            with pytest.raises(DataSourceNoNameException):
                print(f"Testing dwh {dwh_class}")
                data_source = dwh_class(query="test_query")
        else:
            data_source = dwh_class(query="test_query")
            assert data_source.name == ""


@pytest.mark.integration
def test_update_data_sources_with_inferred_event_timestamp_col(simple_dataset_1):
    df_with_two_viable_timestamp_cols = simple_dataset_1.copy(deep=True)
    df_with_two_viable_timestamp_cols["ts_2"] = simple_dataset_1["ts_1"]

    with prep_file_source(df=simple_dataset_1) as file_source:
        data_sources = [
            file_source,
            simple_bq_source_using_table_ref_arg(simple_dataset_1),
            simple_bq_source_using_query_arg(simple_dataset_1),
        ]
        update_data_sources_with_inferred_event_timestamp_col(
            data_sources, RepoConfig(provider="local", project="test")
        )
        actual_event_timestamp_cols = [
            source.event_timestamp_column for source in data_sources
        ]

        assert actual_event_timestamp_cols == ["ts_1", "ts_1", "ts_1"]

    with prep_file_source(df=df_with_two_viable_timestamp_cols) as file_source:
        with pytest.raises(RegistryInferenceFailure):
            # two viable event_timestamp_columns
            update_data_sources_with_inferred_event_timestamp_col(
                [file_source], RepoConfig(provider="local", project="test")
            )


def test_on_demand_features_type_inference():
    # Create Feature Views
    date_request = RequestDataSource(
        name="date_request", schema={"some_date": ValueType.UNIX_TIMESTAMP}
    )

    @on_demand_feature_view(
        inputs={"date_request": date_request},
        features=[
            Feature("output", ValueType.UNIX_TIMESTAMP),
            Feature("string_output", ValueType.STRING),
        ],
    )
    def test_view(features_df: pd.DataFrame) -> pd.DataFrame:
        data = pd.DataFrame()
        data["output"] = features_df["some_date"]
        data["string_output"] = features_df["some_date"].astype(pd.StringDtype())
        return data

    test_view.infer_features()

    @on_demand_feature_view(
        inputs={"date_request": date_request},
        features=[
            Feature("output", ValueType.UNIX_TIMESTAMP),
            Feature("object_output", ValueType.STRING),
        ],
    )
    def invalid_test_view(features_df: pd.DataFrame) -> pd.DataFrame:
        data = pd.DataFrame()
        data["output"] = features_df["some_date"]
        data["object_output"] = features_df["some_date"].astype(str)
        return data

    with pytest.raises(ValueError, match="Value with native type object"):
        invalid_test_view.infer_features()

    @on_demand_feature_view(
        inputs={"date_request": date_request},
        features=[
            Feature("output", ValueType.UNIX_TIMESTAMP),
            Feature("missing", ValueType.STRING),
        ],
    )
    def test_view_with_missing_feature(features_df: pd.DataFrame) -> pd.DataFrame:
        data = pd.DataFrame()
        data["output"] = features_df["some_date"]
        return data

    with pytest.raises(SpecifiedFeaturesNotPresentError):
        test_view_with_missing_feature.infer_features()


def test_datasource_inference():
    # Create Feature Views
    date_request = RequestDataSource(
        name="date_request", schema={"some_date": ValueType.UNIX_TIMESTAMP}
    )

    @on_demand_feature_view(
        inputs={"date_request": date_request},
        features=[
            Feature("output", ValueType.UNIX_TIMESTAMP),
            Feature("string_output", ValueType.STRING),
        ],
    )
    def test_view(features_df: pd.DataFrame) -> pd.DataFrame:
        data = pd.DataFrame()
        data["output"] = features_df["some_date"]
        data["string_output"] = features_df["some_date"].astype(pd.StringDtype())
        return data

    test_view.infer_features()

    @on_demand_feature_view(
        inputs={"date_request": date_request},
        features=[
            Feature("output", ValueType.UNIX_TIMESTAMP),
            Feature("object_output", ValueType.STRING),
        ],
    )
    def invalid_test_view(features_df: pd.DataFrame) -> pd.DataFrame:
        data = pd.DataFrame()
        data["output"] = features_df["some_date"]
        data["object_output"] = features_df["some_date"].astype(str)
        return data

    with pytest.raises(ValueError, match="Value with native type object"):
        invalid_test_view.infer_features()

    @on_demand_feature_view(
        inputs={"date_request": date_request},
        features=[
            Feature("output", ValueType.UNIX_TIMESTAMP),
            Feature("missing", ValueType.STRING),
        ],
    )
    def test_view_with_missing_feature(features_df: pd.DataFrame) -> pd.DataFrame:
        data = pd.DataFrame()
        data["output"] = features_df["some_date"]
        return data

    with pytest.raises(SpecifiedFeaturesNotPresentError):
        test_view_with_missing_feature.infer_features()



import assertpy
import pytest
from google.protobuf.json_format import MessageToDict, Parse

from feast import proto_json
from feast.protos.feast.serving.ServingService_pb2 import (
    FeatureList,
    GetOnlineFeaturesResponse,
)
from feast.protos.feast.types.Value_pb2 import RepeatedValue

FeatureVector = GetOnlineFeaturesResponse.FeatureVector


@pytest.fixture(scope="module")
def proto_json_patch():
    proto_json.patch()


def test_feature_vector_values(proto_json_patch):
    # FeatureVector contains "repeated<feast.types.Value> values" proto field.
    # We want to test that feast.types.Value can take different types in JSON
    # without using additional structure (e.g. 1 instead of {int64_val: 1}).
    feature_vector_str = """{
        "values": [
            1,
            2.0,
            true,
            "foo",
            [1, 2, 3],
            [2.0, 3.0, 4.0, null],
            [true, false, true],
            ["foo", "bar", "foobar"]
        ]
    }"""
    feature_vector_proto = FeatureVector()
    Parse(feature_vector_str, feature_vector_proto)
    assertpy.assert_that(len(feature_vector_proto.values)).is_equal_to(8)
    assertpy.assert_that(feature_vector_proto.values[0].int64_val).is_equal_to(1)
    assertpy.assert_that(feature_vector_proto.values[1].double_val).is_equal_to(2.0)
    assertpy.assert_that(feature_vector_proto.values[2].bool_val).is_equal_to(True)
    assertpy.assert_that(feature_vector_proto.values[3].string_val).is_equal_to("foo")
    assertpy.assert_that(feature_vector_proto.values[4].int64_list_val.val).is_equal_to(
        [1, 2, 3]
    )
    # Can't directly check equality to [2.0, 3.0, 4.0, float("nan")], because float("nan") != float("nan")
    assertpy.assert_that(
        feature_vector_proto.values[5].double_list_val.val[:3]
    ).is_equal_to([2.0, 3.0, 4.0])
    assertpy.assert_that(feature_vector_proto.values[5].double_list_val.val[3]).is_nan()
    assertpy.assert_that(feature_vector_proto.values[6].bool_list_val.val).is_equal_to(
        [True, False, True]
    )
    assertpy.assert_that(
        feature_vector_proto.values[7].string_list_val.val
    ).is_equal_to(["foo", "bar", "foobar"])

    # Now convert protobuf back to json and check that
    feature_vector_json = MessageToDict(feature_vector_proto)
    assertpy.assert_that(len(feature_vector_json["values"])).is_equal_to(8)
    assertpy.assert_that(feature_vector_json["values"][0]).is_equal_to(1)
    assertpy.assert_that(feature_vector_json["values"][1]).is_equal_to(2.0)
    assertpy.assert_that(feature_vector_json["values"][2]).is_equal_to(True)
    assertpy.assert_that(feature_vector_json["values"][3]).is_equal_to("foo")
    assertpy.assert_that(feature_vector_json["values"][4]).is_equal_to([1, 2, 3])
    # Can't directly check equality to [2.0, 3.0, 4.0, float("nan")], because float("nan") != float("nan")
    assertpy.assert_that(feature_vector_json["values"][5][:3]).is_equal_to(
        [2.0, 3.0, 4.0]
    )
    assertpy.assert_that(feature_vector_json["values"][5][3]).is_nan()
    assertpy.assert_that(feature_vector_json["values"][6]).is_equal_to(
        [True, False, True]
    )
    assertpy.assert_that(feature_vector_json["values"][7]).is_equal_to(
        ["foo", "bar", "foobar"]
    )


def test_feast_repeated_value(proto_json_patch):
    # Make sure that RepeatedValue in JSON does not need the
    # additional structure (e.g. [1,2,3] instead of {"val": [1,2,3]})
    repeated_value_str = "[1,2,3]"
    repeated_value_proto = RepeatedValue()
    Parse(repeated_value_str, repeated_value_proto)
    assertpy.assert_that(len(repeated_value_proto.val)).is_equal_to(3)
    assertpy.assert_that(repeated_value_proto.val[0].int64_val).is_equal_to(1)
    assertpy.assert_that(repeated_value_proto.val[1].int64_val).is_equal_to(2)
    assertpy.assert_that(repeated_value_proto.val[2].int64_val).is_equal_to(3)
    # Now convert protobuf back to json and check that
    repeated_value_json = MessageToDict(repeated_value_proto)
    assertpy.assert_that(repeated_value_json).is_equal_to([1, 2, 3])


def test_feature_list(proto_json_patch):
    # Make sure that FeatureList in JSON does not need the additional structure
    # (e.g. ["foo", "bar"] instead of {"val": ["foo", "bar"]})
    feature_list_str = '["feature-a", "feature-b", "feature-c"]'
    feature_list_proto = FeatureList()
    Parse(feature_list_str, feature_list_proto)
    assertpy.assert_that(len(feature_list_proto.val)).is_equal_to(3)
    assertpy.assert_that(feature_list_proto.val[0]).is_equal_to("feature-a")
    assertpy.assert_that(feature_list_proto.val[1]).is_equal_to("feature-b")
    assertpy.assert_that(feature_list_proto.val[2]).is_equal_to("feature-c")
    # Now convert protobuf back to json and check that
    feature_list_json = MessageToDict(feature_list_proto)
    assertpy.assert_that(feature_list_json).is_equal_to(
        ["feature-a", "feature-b", "feature-c"]
    )
